from fedot.core.optimisers.timer import Timer


class TunerTimer(Timer):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def is_time_limit_reached(self) -> bool:
        super().is_time_limit_reached()
        if self.process_terminated:
            self.log.info('Tuning completed because of the time limit reached')
        return self.process_terminated

    def __exit__(self, *args):
        return self.process_terminated



from os.path import join

from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.repository.tasks import Task, TaskTypesEnum
from fedot.core.utils import fedot_project_root


def get_scoring_data():
    file_path_train = join('cases', 'data', 'scoring', 'scoring_train.csv')
    full_path_train = join(str(fedot_project_root()), file_path_train)

    # a dataset for a final validation of the composed model
    file_path_test = join('cases', 'data', 'scoring', 'scoring_test.csv')
    full_path_test = join(str(fedot_project_root()), file_path_test)
    task = Task(TaskTypesEnum.classification)
    train = InputData.from_csv(full_path_train, task=task)
    test = InputData.from_csv(full_path_test, task=task)

    return train, test


def get_kc2_data():
    file_path = join('cases', 'data', 'kc2', 'kc2.csv')
    full_path = join(str(fedot_project_root()), file_path)
    task = Task(TaskTypesEnum.classification)
    data = InputData.from_csv(full_path, task=task)
    train, test = train_test_data_setup(data)

    return train, test


def get_cholesterol_data():
    file_path = join('cases', 'data', 'cholesterol', 'cholesterol.csv')
    full_path = join(str(fedot_project_root()), file_path)
    task = Task(TaskTypesEnum.regression)
    data = InputData.from_csv(full_path, task=task)
    train, test = train_test_data_setup(data)

    return train, test



import logging

from fedot.api.main import Fedot
from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.repository.tasks import TaskTypesEnum, Task
from fedot.core.utils import fedot_project_root


def run_regression_example(visualise=False):
    data_path = f'{fedot_project_root()}/cases/data/cholesterol/cholesterol.csv'

    data = InputData.from_csv(data_path,
                              task=Task(TaskTypesEnum.regression))
    train, test = train_test_data_setup(data)
    problem = 'regression'

    composer_params = {'history_folder': 'custom_history_folder', 'preset': 'auto'}
    auto_model = Fedot(problem=problem, seed=42, timeout=2, logging_level=logging.INFO,
                       **composer_params)

    auto_model.fit(features=train, target='target')
    prediction = auto_model.predict(features=test)
    if visualise:
        auto_model.history.save('saved_regression_history.json')
        auto_model.plot_prediction()
    print(auto_model.get_metrics())
    return prediction


if __name__ == '__main__':
    run_regression_example(visualise=True)



import json
from os.path import join
from typing import List, Optional, Type, Sequence

import numpy as np
from matplotlib import pyplot as plt

from fedot.core.dag.convert import graph_structure_as_nx_graph
from fedot.core.data.data import InputData
from fedot.core.log import default_log
from fedot.core.pipelines.node import Node
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.utils import default_fedot_data_dir
from fedot.sensitivity.node_sa_approaches import NodeAnalysis, NodeAnalyzeApproach
from fedot.sensitivity.sa_requirements import SensitivityAnalysisRequirements


class NodesAnalysis:
    """This class is for nodes sensitivity analysis within a :obj:`Pipeline`\n
    It takes nodes and approaches to be applied to chosen nodes.\n
    To define which nodes to analyze pass them to ``nodes_to_analyze`` filed
    or all nodes will be analyzed.

    Args:
        pipeline: pipeline object to analyze
        train_data: data used for Pipeline training
        test_data: data used for Pipeline validation
        approaches: methods applied to nodes to modify the pipeline or analyze certain operations.\n
            Default: [:obj:`NodeDeletionAnalyze`, :obj:`NodeReplaceOperationAnalyze`]
        nodes_to_analyze: nodes to analyze. Default: all nodes
        path_to_save: path to save results to. Default: ``~home/Fedot/sensitivity``
    """

    def __init__(self, pipeline: Pipeline, train_data: InputData, test_data: InputData,
                 approaches: Optional[List[Type[NodeAnalyzeApproach]]] = None,
                 requirements: SensitivityAnalysisRequirements = None,
                 path_to_save=None, nodes_to_analyze: List[Node] = None):

        self.pipeline = pipeline
        self.train_data = train_data
        self.test_data = test_data
        self.approaches = approaches
        self.requirements = \
            SensitivityAnalysisRequirements() if requirements is None else requirements
        self.metric = self.requirements.metric
        self.log = default_log(self)
        self.path_to_save = \
            join(default_fedot_data_dir(), 'sensitivity', 'nodes_sensitivity') if path_to_save is None else path_to_save

        if not nodes_to_analyze:
            self.log.info('Nodes to analyze are not defined. All nodes will be analyzed.')
            self.nodes_to_analyze = self.pipeline.nodes
        else:
            self.nodes_to_analyze = nodes_to_analyze

    def analyze(self) -> dict:
        """Main method to run the analyze process for every node.

        Returns:
            dict: with analysis result per Node
        """

        nodes_results = dict()
        operation_types = []
        for node in self.nodes_to_analyze:
            node_result = NodeAnalysis(approaches=self.approaches,
                                       approaches_requirements=self.requirements,
                                       path_to_save=self.path_to_save). \
                analyze(pipeline=self.pipeline, node=node,
                        train_data=self.train_data,
                        test_data=self.test_data)
            operation_types.append(node.operation.operation_type)

            nodes_results[f'id = {self.pipeline.nodes.index(node)}, ' \
                          f'operation = {node.content["name"].operation_type}'] = node_result

        if self.requirements.visualization:
            self._visualize_result_per_approach(nodes_results, operation_types)

        if len(self.nodes_to_analyze) == len(self.pipeline.nodes):
            self._visualize_degree_correlation(nodes_results)

        if self.requirements.is_save:
            self._save_results_to_json(nodes_results)
        return nodes_results

    def _save_results_to_json(self, result: dict):
        result_file = join(self.path_to_save, 'nodes_SA_results.json')
        with open(result_file, 'w', encoding='utf-8') as file:
            file.write(json.dumps(result, indent=4))

        self.log.info(f'Pipeline Sensitivity Analysis results were saved to {result_file}')

    def _visualize_result_per_approach(self, results: dict, types: list):
        gathered_results = self._extract_result_values(results)

        for index, result in enumerate(gathered_results):
            colors = ['r' if y < 0 else 'g' for y in result]
            fig, ax = plt.subplots(figsize=(15, 10))
            ax.bar(range(len(results)), result, width=1.0, edgecolor='black', bottom=1,
                   color=colors)
            ax.hlines(1, -1, len(types) + 1, linestyle='--')
            ax.set_xticks(range(len(results)))
            ax.set_xticklabels(types, rotation=45)
            plt.xlabel('iteration')
            plt.ylabel('quality (changed_pipeline_metric/original_metric) - 1')

            file_path = join(self.path_to_save,
                             f'{self.approaches[index].__name__}.jpg')

            plt.savefig(file_path)
            self.log.info(f'Pipeline Sensitivity Analysis visualized results were saved to {file_path}')

    def _visualize_degree_correlation(self, results: dict):
        nodes_degrees = get_nodes_degrees(self.pipeline)
        gathered_results = self._extract_result_values(results)
        for index, result in enumerate(gathered_results):
            fig, ax = plt.subplots(figsize=(15, 10))
            ax.scatter(nodes_degrees, result)

            file_path = join(self.path_to_save,
                             f'{self.approaches[index].__name__}_cor.jpg')
            plt.savefig(file_path)
            self.log.info(f'Nodes degree correlation visualized results were saved to {file_path}')

    def _extract_result_values(self, results):
        gathered_results = []
        for approach in self.approaches:
            approach_result = [np.mean(result[f'{approach.__name__}']) - 1 for result in results.values()]
            gathered_results.append(approach_result)

        return gathered_results


def get_nodes_degrees(graph: 'Graph') -> Sequence[int]:
    """Nodes degree as the number of edges the node has:
        ``degree = #input_edges + #out_edges``

    Returns:
        nodes degrees ordered according to the nx_graph representation of this graph
    """
    graph, _ = graph_structure_as_nx_graph(graph)
    index_degree_pairs = graph.degree
    node_degrees = [node_degree[1] for node_degree in index_degree_pairs]
    return node_degrees



import os
from datetime import timedelta

import numpy as np
import pandas as pd

from fedot.core.data.data import InputData
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.data.supplementary_data import SupplementaryData
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.tuning.tuner_builder import TunerBuilder
from fedot.core.pipelines.tuning.unified import PipelineTuner
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.quality_metrics_repository import RegressionMetricsEnum
from fedot.core.repository.tasks import TaskTypesEnum, Task
from test.unit.api.test_api_cli_params import project_root_path


def get_multitask_pipeline():
    logit_node = PrimaryNode('logit')
    data_source_node = PrimaryNode('data_source_table/regression')
    final_node = SecondaryNode('dtreg', nodes_from=[logit_node, data_source_node])
    return Pipeline(final_node)


def prepare_multitask_data() -> (MultiModalData, MultiModalData):
    """ Load data for multitask regression / classification pipeline """
    ex_data = os.path.join(project_root_path, 'examples/data')
    train_df = pd.read_csv(os.path.join(ex_data, 'train_synthetic_regression_classification.csv'))
    test_df = pd.read_csv(os.path.join(ex_data, 'test_synthetic_regression_classification.csv'))

    # Data for classification
    class_task = Task(TaskTypesEnum.classification)
    class_train = InputData(idx=np.arange(0, len(train_df)), features=np.array(train_df[['feature_1', 'feature_2']]),
                            target=np.array(train_df['class']), task=class_task, data_type=DataTypesEnum.table,
                            supplementary_data=SupplementaryData(is_main_target=False))
    class_test = InputData(idx=np.arange(0, len(test_df)), features=np.array(test_df[['feature_1', 'feature_2']]),
                           target=None, task=class_task, data_type=DataTypesEnum.table,
                           supplementary_data=SupplementaryData(is_main_target=False))

    # Data for regression
    task = Task(TaskTypesEnum.regression)
    reg_train = InputData(idx=np.arange(0, len(train_df)), features=np.array(train_df[['feature_1', 'feature_2']]),
                          target=np.array(train_df['concentration']), task=task, data_type=DataTypesEnum.table)
    reg_test = InputData(idx=np.arange(0, len(test_df)), features=np.array(test_df[['feature_1', 'feature_2']]),
                         target=None, task=task, data_type=DataTypesEnum.table)

    train_multimodal = MultiModalData({'logit': class_train,
                                       'data_source_table/regression': reg_train})

    test_multimodal = MultiModalData({'logit': class_test,
                                      'data_source_table/regression': reg_test})
    return train_multimodal, test_multimodal


def launch_multitask_example(with_tuning: bool = False):
    """
    Demonstration of an example with running a multitask pipeline.
    Synthetic data is used. Task: predict the category of the substance (column "class") <- classification,
    and then predict the concentration based on the predicted category (column "concentration") <- regression.

    :param with_tuning: is tuning required or not
    """
    train_input, test_input = prepare_multitask_data()
    multitask_pipeline = get_multitask_pipeline()

    if with_tuning:
        tuner = TunerBuilder(train_input.task)\
            .with_tuner(PipelineTuner)\
            .with_metric(RegressionMetricsEnum.MAE)\
            .with_iterations(100)\
            .with_timeout(timedelta(minutes=2))\
            .build(train_input)
        multitask_pipeline = tuner.tune(multitask_pipeline)

    multitask_pipeline.fit(train_input)
    side_pipeline = multitask_pipeline.pipeline_for_side_task(task_type=TaskTypesEnum.classification)

    # Replace the name of main "data source" in preprocessor
    side_pipeline.preprocessor.main_target_source_name = 'logit'
    side_output = side_pipeline.predict(test_input, output_mode='labels')
    output = multitask_pipeline.predict(test_input)

    print(f'Predicted classes: {np.ravel(side_output.predict)}')
    print(f'Predicted concentrations: {np.ravel(output.predict)}')


if __name__ == '__main__':
    launch_multitask_example(with_tuning=True)



import os
import timeit
import warnings

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from sklearn.metrics import mean_squared_error, mean_absolute_error

from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TsForecastingParams, TaskTypesEnum
from fedot.core.utils import fedot_project_root

warnings.filterwarnings('ignore')
np.random.seed(2020)


def make_forecast(pipeline, train: InputData, predict: InputData,
                  train_exog: InputData, predict_exog: InputData):
    """
    Function for predicting values in a time series

    :return predicted_values: numpy array, forecast of model
    """

    # Fit it
    start_time = timeit.default_timer()

    second_node_name = 'exog_ts'

    if train_exog is None:
        second_node_name = 'data_source_ts/2'
        train_exog = train
        predict_exog = predict

    train_dataset = MultiModalData({
        'data_source_ts/1': train,
        second_node_name: train_exog})

    predict_dataset = MultiModalData({
        'data_source_ts/1': predict,
        second_node_name: predict_exog})

    pipeline.fit_from_scratch(train_dataset)
    amount_of_seconds = timeit.default_timer() - start_time

    print(f'\nIt takes {amount_of_seconds:.2f} seconds to train pipeline\n')

    # Predict
    predicted_values = pipeline.predict(predict_dataset)
    predicted_values = predicted_values.predict

    return predicted_values


def run_exogenous_experiment(path_to_file, len_forecast=250, with_exog=True,
                             visualization=False) -> None:
    """ Function with example how time series forecasting can be made with using
    exogenous features

    :param path_to_file: path to the csv file with dataframe
    :param len_forecast: forecast length
    :param with_exog: is it needed to make prediction with exogenous time series
    :param visualization: is it needed to make visualizations
    """

    df = pd.read_csv(path_to_file)
    time_series = np.array(df['Level'])
    exog_variable = np.array(df['Neighboring level'])

    # Source time series
    train_input, predict_input = train_test_data_setup(InputData(idx=range(len(time_series)),
                                                                 features=time_series,
                                                                 target=time_series,
                                                                 task=Task(TaskTypesEnum.ts_forecasting,
                                                                           TsForecastingParams(
                                                                               forecast_length=len_forecast)),
                                                                 data_type=DataTypesEnum.ts))

    # Exogenous time series
    task = Task(TaskTypesEnum.ts_forecasting, TsForecastingParams(forecast_length=len_forecast))

    predict_input_exog = InputData(idx=np.arange(len(exog_variable)),
                                   features=exog_variable, target=time_series,
                                   task=task, data_type=DataTypesEnum.ts)

    train_input_exog, predict_input_exog = train_test_data_setup(predict_input_exog)

    if with_exog is True:
        # Example with exogenous time series
        node_source = PrimaryNode('data_source_ts/1')
        node_lagged = SecondaryNode('lagged', nodes_from=[node_source])

        node_exog = PrimaryNode('exog_ts')

        node_final = SecondaryNode('ridge', nodes_from=[node_lagged, node_exog])
        pipeline = Pipeline(node_final)
    else:
        # Simple example without exogenous time series
        node_source_1 = PrimaryNode('data_source_ts/1')
        node_source_2 = PrimaryNode('data_source_ts/2')

        node_lagged_1 = SecondaryNode('lagged', nodes_from=[node_source_1])
        node_lagged_2 = SecondaryNode('lagged', nodes_from=[node_source_2])

        node_ridge_1 = SecondaryNode('ridge', nodes_from=[node_lagged_1])
        node_ridge_2 = SecondaryNode('ridge', nodes_from=[node_lagged_2])
        node_final = SecondaryNode('ridge', nodes_from=[node_ridge_1, node_ridge_2])
        train_input_exog = None
        predict_input_exog = None
        pipeline = Pipeline(node_final)

    predicted = make_forecast(pipeline, train_input, predict_input, train_input_exog, predict_input_exog)

    predicted = np.ravel(np.array(predicted))
    test_data = np.ravel(predict_input.target)

    print(f'Predicted values: {predicted[:5]}')
    print(f'Actual values: {test_data[:5]}')

    mse_before = mean_squared_error(test_data, predicted, squared=False)
    mae_before = mean_absolute_error(test_data, predicted)
    print(f'RMSE - {mse_before:.4f}')
    print(f'MAE - {mae_before:.4f}\n')

    if visualization:
        plt.plot(range(0, len(time_series)), time_series, label='Actual time series')
        plt.plot(range(len(train_input.target), len(time_series)), predicted, label='Forecast')
        plt.legend()
        plt.grid()
        plt.show()


if __name__ == '__main__':
    data_path = os.path.join(f'{fedot_project_root()}', 'examples/data/ts', 'ts_sea_level.csv')
    run_exogenous_experiment(path_to_file=data_path, len_forecast=250, with_exog=True, visualization=True)



import math
from copy import copy
from typing import Union, Optional

import numpy as np

from fedot.core.data.data import InputData, OutputData
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.operations.evaluation.operation_implementations.data_operations.ts_transformations import ts_to_table
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import TaskTypesEnum


def out_of_sample_ts_forecast(pipeline: Pipeline, input_data: Union[InputData, MultiModalData],
                              horizon: int = None) -> np.array:
    """
    Method allow make forecast with appropriate forecast length. The previously
    predicted parts of the time series are used for forecasting next parts. Available
    only for time series forecasting task. Steps ahead provided iteratively.
    time series ----------------|
    forecast                    |---|---|---|

    :param pipeline: Pipeline for making time series forecasting
    :param input_data: data for prediction
    :param horizon: forecasting horizon
    :return final_forecast: array with forecast
    """
    task = input_data.task
    exception_if_not_ts_task(task)

    # How many elements to the future pipeline can produce
    forecast_length = task.task_params.forecast_length

    final_forecast = []
    if isinstance(input_data, MultiModalData) or input_data.data_type == DataTypesEnum.multi_ts:
        if forecast_length < horizon:
            raise ValueError('In case of multi-modal time-series and multi time-series'
                             'forecast horizon can not be bigger than forecast length model was fitted for.\n'
                             f'forecast_length = {forecast_length}\n'
                             f'horizon = {horizon}')
        iter_predict = pipeline.predict(input_data).predict
        iter_predict = np.ravel(np.array(iter_predict))
        final_forecast.append(iter_predict)
    else:
        pre_history_ts = np.array(input_data.features)
        number_of_iterations = math.ceil(horizon / forecast_length)
        # Make forecast iteratively moving throw the horizon
        for _ in range(0, number_of_iterations):
            iter_predict = pipeline.predict(input_data=input_data).predict
            iter_predict = np.ravel(np.array(iter_predict))
            final_forecast.append(iter_predict)

            # Add prediction to the historical data - update it
            pre_history_ts = np.hstack((pre_history_ts, iter_predict))

            # Prepare InputData for next iteration
            input_data = _update_input(pre_history_ts, forecast_length, task)

    # Create output data
    final_forecast = np.ravel(np.array(final_forecast))
    if final_forecast.ndim > 1:
        final_forecast = np.squeeze(final_forecast)
    # Clip the forecast if it is necessary
    final_forecast = final_forecast[:horizon]
    return final_forecast


def in_sample_ts_forecast(pipeline, input_data: Union[InputData, MultiModalData],
                          horizon: int = None) -> np.array:
    """
    Method allows to make in-sample forecasting. The actual values of the time
    series, rather than the previously predicted parts of the time series,
    are used for forecasting next parts. Sources indices in input_data parameter
    will be ignored
    time series ----------------|---|---|---|
    forecast                    |---|---|---|

    :param pipeline: Pipeline for making time series forecasting
    :param input_data: data for prediction
    :param horizon: forecasting horizon
    :return final_forecast: array with forecast
    """
    # Divide data on samples into pre-history and validation part
    task = input_data.task
    exception_if_not_ts_task(task)
    # How many elements to the future pipeline can produce
    scope_len = task.task_params.forecast_length
    number_of_iterations = _calculate_number_of_steps(scope_len, horizon)

    if isinstance(input_data, InputData):
        time_series = np.array(input_data.features)
        pre_history_ts = time_series[:-horizon]
        source_len = len(pre_history_ts)
        last_index_pre_history = source_len - 1

        data = _update_input(pre_history_ts, scope_len, task, input_data.data_type)
    else:
        # TODO simplify

        data = MultiModalData()
        for data_id in input_data.keys():
            features = input_data[data_id].features
            time_series = np.array(features)
            pre_history_ts = time_series[:-horizon]
            source_len = len(pre_history_ts)
            last_index_pre_history = source_len - 1

            local_data = _update_input(pre_history_ts, scope_len, task, input_data[data_id].data_type)
            data[data_id] = local_data

    # Calculate intervals
    intervals = _calculate_intervals(last_index_pre_history,
                                     number_of_iterations,
                                     scope_len)

    # Make forecast iteratively moving throw the horizon
    final_forecast = []
    for _, border in zip(range(0, number_of_iterations), intervals):

        iter_predict = pipeline.predict(input_data=data)
        iter_predict = np.ravel(np.array(iter_predict.predict))
        final_forecast.append(iter_predict)

        if isinstance(input_data, InputData):
            # Add actual values to the historical data - update it
            pre_history_ts = time_series[:border + 1]
            # Prepare InputData for next iteration
            data = _update_input(pre_history_ts, scope_len, task, input_data.data_type)
        else:
            # TODO simplify
            data = MultiModalData()
            for data_id in input_data.keys():
                features = input_data[data_id].features
                time_series = np.array(features)
                pre_history_ts = time_series[:border + 1]
                local_data = _update_input(pre_history_ts, scope_len, task, input_data[data_id].data_type)
                data[data_id] = local_data

    # Create output data
    final_forecast = np.ravel(np.array(final_forecast))
    # Clip the forecast if it is necessary
    final_forecast = final_forecast[:horizon]
    return final_forecast


def fitted_values(source_input: InputData, train_predicted: OutputData, horizon_step: int = None) -> OutputData:
    """ The method converts a multidimensional lagged array into an
    one-dimensional array - time series based on predicted values for training sample

    :param source_input: InputData which were used to train pipeline
    :param train_predicted: OutputData with trained values
    :param horizon_step: index of elements for forecast. If None - perform
    averaging for all forecasting steps
    """
    copied_data = copy(train_predicted)
    if horizon_step is not None:
        # Take particular forecast step
        copied_data.predict = copied_data.predict[:, horizon_step]
        if isinstance(copied_data.idx, list):
            # if indices can not be incremented, replace it
            copied_data.idx = generate_ids(source_input, copied_data, expand=False)
        copied_data.idx = copied_data.idx + horizon_step
        return copied_data
    else:
        # Perform collapse with averaging
        forecast_length = copied_data.task.task_params.forecast_length

        # Extend source index range
        if isinstance(copied_data.idx, list):
            indices_range = generate_ids(source_input, copied_data, expand=True)
        else:
            # if indices can be incremented
            indices_range = np.arange(copied_data.idx[0],
                                      copied_data.idx[-1] + forecast_length)

        # Lagged matrix with indices in cells
        _, idx_matrix = ts_to_table(idx=indices_range,
                                    time_series=indices_range,
                                    window_size=forecast_length)
        predicted_matrix = copied_data.predict

        # For every index calculate mean predictions (by all forecast steps)
        final_predictions = []
        for index in indices_range:
            vals = predicted_matrix[idx_matrix == index]
            mean_value = np.mean(vals)
            final_predictions.append(mean_value)
        copied_data.predict = np.array(final_predictions)
        copied_data.idx = indices_range
        return copied_data


def in_sample_fitted_values(source_input: InputData, train_predicted: OutputData) -> OutputData:
    """ Perform in sample validation based on training sample """
    forecast_length = train_predicted.task.task_params.forecast_length
    all_values = []
    step = 0
    # Glues together parts of predictions using "in-sample" way
    while step < len(train_predicted.predict):
        all_values.extend(train_predicted.predict[step, :])
        step += forecast_length

    # In some cases it doesn't reach the end
    if not np.isclose(all_values[-1], train_predicted.predict[-1, -1]):
        missing_part_index = step - len(train_predicted.predict) + 1
        # Store missing predicted values
        all_values.extend(train_predicted.predict[-1, missing_part_index:])

    copied_data = copy(train_predicted)
    copied_data.predict = np.array(all_values)
    # Update indices
    first_id = copied_data.idx[0]
    if isinstance(first_id, str):
        indices_range = generate_ids(source_input, copied_data, expand=True)
        copied_data.idx = indices_range[:-1]
    else:
        copied_data.idx = np.arange(first_id, first_id + len(all_values))

    return copied_data


def _calculate_number_of_steps(scope_len, horizon):
    """ Method return amount of iterations which must be done for multistep
    time series forecasting

    :param scope_len: time series forecasting length
    :param horizon: forecast horizon
    :return amount_of_steps: amount of steps to produce
    """
    amount_of_iterations = int(horizon // scope_len)

    # Remainder of the division
    resid = int(horizon % scope_len)
    if resid == 0:
        amount_of_steps = amount_of_iterations
    else:
        amount_of_steps = amount_of_iterations + 1

    return amount_of_steps


def _update_input(pre_history_ts, scope_len, task, data_type: DataTypesEnum = DataTypesEnum.ts):
    """ Method make new InputData object based on the previous part of time
    series

    Args:
        pre_history_ts: time series
        scope_len: how many elements to the future can algorithm forecast
        task: time series forecasting task
        data_type: type of data (ts or multi_ts

    Returns:
        input_data: updated InputData
    """
    start_forecast = len(pre_history_ts)
    end_forecast = start_forecast + scope_len
    input_data = InputData(idx=np.arange(start_forecast, end_forecast),
                           features=pre_history_ts, target=None,
                           task=task, data_type=data_type)

    return input_data


def _calculate_intervals(last_index_pre_history, amount_of_iterations, scope_len):
    """ Function calculate

    :param last_index_pre_history: last id of the known part of time series
    :param amount_of_iterations: amount of steps for time series forecasting
    :param scope_len: amount of elements in every time series forecasting step
    :return intervals: ids of finish of every step in time series
    """
    intervals = []
    current_border = last_index_pre_history
    for i in range(0, amount_of_iterations):
        current_border = current_border + scope_len
        intervals.append(current_border)

    return intervals


def exception_if_not_ts_task(task):
    if task.task_type is not TaskTypesEnum.ts_forecasting:
        raise ValueError('Method forecast is available only for time series forecasting task')


def generate_ids(source_input, output_data, expand: bool):
    """
    Create new indices for fitted time series values. Process shift after
    starting clipping.

    :param source_input: source idx, features and target
    :param output_data: OutputData after pipeline fit
    :param expand: there is a need to take into account the forecasting horizon.
    If True, expand indices.
    """
    forecast_len = source_input.task.task_params.forecast_length
    source_idx = source_input.idx[:-forecast_len]

    # Calculate difference between source idx len and new
    clipped_starting_values = len(source_idx) - len(output_data.idx)

    if expand:
        indices_range = np.arange(clipped_starting_values + 1,
                                  len(source_idx) + forecast_len + 1)
    else:
        indices_range = np.arange(clipped_starting_values + 1,
                                  len(source_idx) + 1)
    return indices_range


def convert_forecast_to_output(pre_history_data: Union[InputData, MultiModalData], forecast: np.array,
                               idx: Optional[np.array] = None) -> OutputData:
    """Converts forecast array to OutputData

    Args:
        pre_history_data: data which was used for prediction
        forecast: array with predicted values
        idx: array with idx values. If None sets next idx after `pre_history_data` with length of forecast for InputData
        and idx from 0 to forecast length for MultimodalData.
    """
    features = pre_history_data.features if isinstance(pre_history_data, InputData) else None
    if forecast.ndim > 1:
        forecast = np.squeeze(forecast)
    if idx is None:
        if features is not None:
            idx = np.arange(len(features), len(features) + len(forecast))
        else:
            idx = np.arange(len(forecast))
    prediction = OutputData(idx=idx,
                            features=features,
                            predict=forecast,
                            task=pre_history_data.task,
                            data_type=DataTypesEnum.ts)
    return prediction



from typing import Optional, Union

from fedot.core.data.data import InputData
from fedot.core.data.data_preprocessing import data_has_missing_values, data_has_categorical_features, \
    data_has_text_features
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.pipelines.node import Node
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.pipeline_builder import PipelineBuilder
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import TaskTypesEnum


class PreprocessingBuilder:
    """
    Builder for constructing preprocessing part of pipeline during the preparation of an initial assumption.
    If data is multimodal, builder makes preprocessing pipeline for each data source iteratively.
    """
    def __init__(self, task_type: TaskTypesEnum, data_type: DataTypesEnum, *initial_nodes: Node):
        self.task_type = task_type
        self.data_type = data_type
        self._builder = PipelineBuilder(*initial_nodes)

    @classmethod
    def builder_for_data(cls,
                         task_type: TaskTypesEnum,
                         data: Union[InputData, MultiModalData],
                         *initial_nodes: Optional[Node]) -> PipelineBuilder:
        if isinstance(data, MultiModalData):
            # if the data is unimodal, initial_nodes = tuple of None
            # if the data is multimodal, initial_nodes = tuple of 1 element (current data_source node)
            # so the whole data is reduced to the current data_source for an easier preprocessing
            data = data[str(initial_nodes[0])]
        preprocessing_builder = cls(task_type, data.data_type, *initial_nodes)
        if data_has_text_features(data):
            preprocessing_builder = preprocessing_builder.with_text_vectorizer()
        return preprocessing_builder.to_builder()

    def with_scaling(self):
        if self.task_type is not TaskTypesEnum.ts_forecasting and self.data_type is not DataTypesEnum.image:
            self._builder.add_node('scaling')
        return self

    def with_text_vectorizer(self):
        self._builder.add_node('tfidf')
        return self

    def to_builder(self) -> PipelineBuilder:
        """ Return result as PipelineBuilder. Scaling is applied final by default. """
        return self.with_scaling()._builder

    def to_pipeline(self) -> Optional[Pipeline]:
        """ Return result as Pipeline. Scaling is applied final by default. """
        return self.to_builder().to_pipeline()



from copy import copy

import numpy as np
import statsmodels.api as sm
from statsmodels.genmod.families import Gamma, Gaussian, InverseGaussian
from statsmodels.genmod.families.links import identity, inverse_power, inverse_squared, log as lg
from statsmodels.genmod.generalized_linear_model import GLM
from statsmodels.tsa.ar_model import AutoReg
from statsmodels.tsa.exponential_smoothing.ets import ETSModel

from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.operation_implementations.data_operations.ts_transformations import ts_to_table
from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import ModelImplementation
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum


class GLMImplementation(ModelImplementation):
    """ Generalized linear models implementation """
    # some models are dropped due to instability
    family_distribution = {
        "gaussian": {'distribution': Gaussian,
                     'default_link': 'identity',
                     'available_links': {'log': lg(),
                                         'identity': identity(),
                                         'inverse_power': inverse_power()
                                         }
                     },
        "gamma": {'distribution': Gamma,
                  'default_link': 'inverse_power',
                  'available_links': {'log': lg(),
                                      'identity': identity(),
                                      'inverse_power': inverse_power()
                                      }
                  },
        "inverse_gaussian": {'distribution': InverseGaussian,
                             'default_link': 'inverse_squared',
                             'available_links': {'log': lg(),
                                                 'identity': identity(),
                                                 'inverse_squared': inverse_squared(),
                                                 'inverse_power': inverse_power()
                                                 }
                             },
        "default": Gaussian(identity())
    }

    def __init__(self, params: OperationParameters):
        super().__init__(params)
        self.model = None
        self.family_link = None

        self.correct_params()

    @property
    def family(self) -> str:
        return self.params.get('family')

    @property
    def link(self) -> str:
        return self.params.get('link')

    def fit(self, input_data):
        self.model = GLM(
            exog=sm.add_constant(input_data.idx.astype("float64")).reshape(-1, 2),
            endog=input_data.target.astype("float64").reshape(-1, 1),
            family=self.family_link
        ).fit(method="lbfgs")
        return self.model

    def predict(self, input_data):
        input_data = copy(input_data)
        parameters = input_data.task.task_params
        forecast_length = parameters.forecast_length
        old_idx = input_data.idx
        if forecast_length == 1:
            predictions = self.model.predict(np.concatenate([np.array([1]),
                                                             input_data.idx.astype("float64")]).reshape(-1, 2))
        else:
            predictions = self.model.predict(sm.add_constant(input_data.idx.astype("float64")).reshape(-1, 2))

        start_id = old_idx[-1] - forecast_length + 1
        end_id = old_idx[-1]
        predict = predictions
        predict = np.array(predict).reshape(1, -1)
        new_idx = np.arange(start_id, end_id + 1)

        input_data.idx = new_idx

        output_data = self._convert_to_output(input_data,
                                              predict=predict,
                                              data_type=DataTypesEnum.table)
        return output_data

    def predict_for_fit(self, input_data: InputData) -> OutputData:
        input_data = copy(input_data)
        parameters = input_data.task.task_params
        forecast_length = parameters.forecast_length
        old_idx = input_data.idx
        target = input_data.target
        predictions = self.model.predict(sm.add_constant(input_data.idx.astype("float64")).reshape(-1, 2))
        _, predict = ts_to_table(idx=old_idx,
                                 time_series=predictions,
                                 window_size=forecast_length)
        new_idx, target_columns = ts_to_table(idx=old_idx,
                                              time_series=target,
                                              window_size=forecast_length)

        input_data.idx = new_idx
        input_data.target = target_columns

        output_data = self._convert_to_output(input_data,
                                              predict=predict,
                                              data_type=DataTypesEnum.table)
        return output_data

    def set_default(self):
        """ Set default value of Family(link) """
        self.family_link = self.family_distribution['default']
        self.params.update(family='gaussian')
        self.log.info("Invalid family. Changed to default value")

    def correct_params(self):
        """ Correct params if they are not correct """
        if self.family in self.family_distribution:
            if self.link not in self.family_distribution[self.family]['available_links']:
                # get default link for distribution if current invalid
                default_link = self.family_distribution[self.family]['default_link']
                self.log.info(
                    f"Invalid link function {self.link} for {self.family}. Change to default "
                    f"link {default_link}")
                self.params.update(link=default_link)
            # if correct isn't need
            self.family_link = self.family_distribution[self.family]['distribution'](
                self.family_distribution[self.family]['available_links'][self.link]
            )
        else:
            # get default family for distribution if current invalid
            self.set_default()


class AutoRegImplementation(ModelImplementation):

    def __init__(self, params: OperationParameters):
        super().__init__(params)
        self.autoreg = None
        self.actual_ts_len = None

    def fit(self, input_data):
        """ Class fit ar model on data

        :param input_data: data with features, target and ids to process
        """

        source_ts = np.array(input_data.features)
        self.actual_ts_len = len(source_ts)

        # Correct window size parameter
        self._check_and_correct_lags(source_ts)

        lag_1 = int(self.params.get('lag_1'))
        lag_2 = int(self.params.get('lag_2'))
        self.autoreg = AutoReg(source_ts, lags=[lag_1, lag_2]).fit()
        self.actual_ts_len = input_data.idx.shape[0]

        return self.autoreg

    def predict(self, input_data):
        """ Method for time series prediction on forecast length

        :param input_data: data with features, target and ids to process
        :return output_data: output data with smoothed time series
        """
        input_data = copy(input_data)
        parameters = input_data.task.task_params
        forecast_length = parameters.forecast_length

        # in case in(out) sample forecasting
        self.handle_new_data(input_data)
        start_id = self.actual_ts_len
        end_id = start_id + forecast_length - 1
        predicted = self.autoreg.predict(start=start_id, end=end_id)
        predict = np.array(predicted).reshape(1, -1)

        output_data = self._convert_to_output(input_data,
                                              predict=predict,
                                              data_type=DataTypesEnum.table)
        return output_data

    def predict_for_fit(self, input_data: InputData) -> OutputData:
        input_data = copy(input_data)
        parameters = input_data.task.task_params
        forecast_length = parameters.forecast_length
        idx = input_data.idx
        target = input_data.target
        predicted = self.autoreg.predict(start=idx[0], end=idx[-1])
        # adding nan to target as in predicted
        nan_mask = np.isnan(predicted)
        target = target.astype(float)
        target[nan_mask] = np.nan
        _, predict = ts_to_table(idx=idx,
                                 time_series=predicted,
                                 window_size=forecast_length)
        _, target_columns = ts_to_table(idx=idx,
                                              time_series=target,
                                              window_size=forecast_length)

        input_data.idx = input_data.idx[~nan_mask]
        input_data.target = target_columns
        output_data = self._convert_to_output(input_data,
                                              predict=predict,
                                              data_type=DataTypesEnum.table)
        return output_data

    def _check_and_correct_lags(self, time_series: np.array):
        previous_lag_1 = int(self.params.get('lag_1'))
        previous_lag_2 = int(self.params.get('lag_2'))
        max_lag = len(time_series) // 2 - 1
        new_lag_1 = self._check_and_correct_lag(max_lag, previous_lag_1)
        new_lag_2 = self._check_and_correct_lag(max_lag, previous_lag_2)
        if new_lag_1 == new_lag_2:
            new_lag_2 -= 1
        prefix = "Warning: lag of AutoRegImplementation was changed"
        if previous_lag_1 != new_lag_1:
            self.log.info(f"{prefix} from {previous_lag_1} to {new_lag_1}.")
            self.params.update(lag_1=new_lag_1)
        if previous_lag_2 != new_lag_2:
            self.log.info(f"{prefix} from {previous_lag_2} to {new_lag_2}.")
            self.params.update(lag_2=new_lag_2)

    def _check_and_correct_lag(self, max_lag: int, lag: int):
        if lag > max_lag:
            lag = max_lag
        return lag

    def handle_new_data(self, input_data: InputData):
        """
        Method to update x samples inside a model (used when we want to use old model to a new data)

        :param input_data: new input_data
        """
        if input_data.idx[0] > self.actual_ts_len:
            self.autoreg.model.endog = input_data.features[-self.actual_ts_len:]
            self.autoreg.model._setup_regressors()


class ExpSmoothingImplementation(ModelImplementation):
    """ Exponential smoothing implementation from statsmodels """

    def __init__(self, params: OperationParameters):
        super().__init__(params)
        self.model = None
        if self.params.get("seasonal"):
            self.seasonal_periods = int(self.params.get("seasonal_periods"))
        else:
            self.seasonal_periods = None

    def fit(self, input_data):
        self.model = ETSModel(
            input_data.features.astype("float64"),
            error=self.params.get("error"),
            trend=self.params.get("trend"),
            seasonal=self.params.get("seasonal"),
            damped_trend=self.params.get("damped_trend"),
            seasonal_periods=self.seasonal_periods
        )
        self.model = self.model.fit(disp=False)
        return self.model

    def predict(self, input_data):
        input_data = copy(input_data)
        idx = input_data.idx

        start_id = idx[0]
        end_id = idx[-1]
        predictions = self.model.predict(start=start_id,
                                         end=end_id)
        predict = predictions
        predict = np.array(predict).reshape(1, -1)
        new_idx = np.arange(start_id, end_id + 1)

        input_data.idx = new_idx

        output_data = self._convert_to_output(input_data,
                                              predict=predict,
                                              data_type=DataTypesEnum.table)
        return output_data

    def predict_for_fit(self, input_data: InputData) -> OutputData:
        input_data = copy(input_data)
        parameters = input_data.task.task_params
        forecast_length = parameters.forecast_length
        idx = input_data.idx
        target = input_data.target

        # Indexing for statsmodels is different
        start_id = idx[0]
        end_id = idx[-1]
        predictions = self.model.predict(start=start_id,
                                         end=end_id)
        _, predict = ts_to_table(idx=idx,
                                 time_series=predictions,
                                 window_size=forecast_length)
        new_idx, target_columns = ts_to_table(idx=idx,
                                              time_series=target,
                                              window_size=forecast_length)

        input_data.idx = new_idx
        input_data.target = target_columns

        output_data = self._convert_to_output(input_data,
                                              predict=predict,
                                              data_type=DataTypesEnum.table)
        return output_data



from copy import copy
from typing import Optional

import numpy as np

from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.operation_implementations.data_operations.ts_transformations import ts_to_table, \
    transform_features_and_target_into_lagged
from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import ModelImplementation
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum


class RepeatLastValueImplementation(ModelImplementation):
    """
    Repeat last known value of time series to the future -
    LOCF (last observation carried forward)
    """

    def __init__(self, params: OperationParameters):
        super().__init__(params)
        self.elements_to_repeat = None

    @property
    def part_for_repeat(self):
        """Which part of time series will be used for repeating. Vary from 0.01 to 0.5
        If -1 - repeat only last value"""
        return self.params.get('part_for_repeat')

    def fit(self, input_data):
        """ Determine how many elements to repeat during forecasting """
        if self.part_for_repeat == -1:
            self.elements_to_repeat = 1
            return self

        elements_to_repeat = round(len(input_data.features) * self.part_for_repeat)
        if elements_to_repeat < 1:
            # Minimum number of elements is one
            self.elements_to_repeat = 1
        else:
            self.elements_to_repeat = elements_to_repeat

        if self.elements_to_repeat > input_data.task.task_params.forecast_length:
            self.elements_to_repeat = input_data.task.task_params.forecast_length
        return self

    def predict(self, input_data: InputData) -> OutputData:
        input_data = copy(input_data)
        forecast_length = input_data.task.task_params.forecast_length

        # Get last known value from history
        last_observations = input_data.features[-self.elements_to_repeat:].reshape(1, -1)
        forecast = self._generate_repeated_forecast(last_observations, forecast_length)

        output_data = self._convert_to_output(input_data,
                                              predict=forecast,
                                              data_type=DataTypesEnum.table)
        return output_data

    def predict_for_fit(self, input_data: InputData) -> OutputData:
        input_data = copy(input_data)
        forecast_length = input_data.task.task_params.forecast_length
        # Transform the predicted time series into a table
        new_idx, transformed_cols, new_target = transform_features_and_target_into_lagged(input_data,
                                                                                          forecast_length,
                                                                                          self.elements_to_repeat)
        input_data.idx = new_idx
        input_data.target = new_target
        forecast = self._generate_repeated_forecast(transformed_cols, forecast_length)
        output_data = self._convert_to_output(input_data,
                                              predict=forecast,
                                              data_type=DataTypesEnum.table)
        return output_data

    def _generate_repeated_forecast(self, transformed_cols: np.array, forecast_length: int):
        if self.elements_to_repeat == 1:
            return np.repeat(transformed_cols, forecast_length, axis=1)
        elif self.elements_to_repeat < forecast_length:
            # Generate pattern
            repeat_number = int(forecast_length / self.elements_to_repeat) + 1
            forecast = np.tile(transformed_cols, repeat_number)
            return forecast[:, :forecast_length]
        else:
            # Number of elements to repeat equal to forecast horizon
            return transformed_cols


class NaiveAverageForecastImplementation(ModelImplementation):
    """ Class for forecasting time series with mean """

    def __init__(self, params: OperationParameters):
        super().__init__(params)

    @property
    def part_for_averaging(self):
        return self.params.get('part_for_averaging')

    def fit(self, input_data):
        """ Such a simple approach does not support fit method """
        pass

    def predict(self, input_data: InputData) -> OutputData:
        """ Get desired part of time series for averaging and calculate mean value """
        forecast_length = input_data.task.task_params.forecast_length

        elements_to_take = self._how_many_elements_use_for_averaging(input_data.features)
        # Prepare single forecast
        mean_value = np.nanmean(input_data.features[-elements_to_take:])
        forecast = np.array([mean_value] * forecast_length).reshape((1, -1))

        output_data = self._convert_to_output(input_data,
                                              predict=forecast,
                                              data_type=DataTypesEnum.table)
        return output_data

    def predict_for_fit(self, input_data: InputData) -> OutputData:
        forecast_length = input_data.task.task_params.forecast_length
        parts = split_rolling_slices(input_data)
        mean_values_for_chunks = self.average_by_axis(parts)
        forecast = np.repeat(mean_values_for_chunks.reshape((-1, 1)), forecast_length, axis=1)
        forecast = forecast[:-forecast_length, :]

        # Update target
        _, transformed_target = ts_to_table(idx=input_data.idx, time_series=input_data.target,
                                            window_size=forecast_length, is_lag=True)
        input_data.target = transformed_target[1:, :]

        # Update indices - there is no forecast for first element and skip last out of boundaries predictions
        last_threshold = forecast_length - 1
        new_idx = input_data.idx[1: -last_threshold]
        input_data.idx = new_idx
        output_data = self._convert_to_output(input_data,
                                              predict=forecast,
                                              data_type=DataTypesEnum.table)
        return output_data

    def average_by_axis(self, parts: np.array):
        """ Perform averaging for each column using last part of it """
        mean_values_for_chunks = np.apply_along_axis(self._average, 1, parts)
        return mean_values_for_chunks

    def _average(self, row: np.array):
        row = row[np.logical_not(np.isnan(row))]
        if len(row) == 1:
            return row

        elements_to_take = self._how_many_elements_use_for_averaging(row)
        return np.mean(row[-elements_to_take:])

    def _how_many_elements_use_for_averaging(self, time_series: np.array):
        elements_to_take = round(len(time_series) * self.part_for_averaging)
        elements_to_take = fix_elements_number(elements_to_take)
        return elements_to_take


def split_rolling_slices(input_data: InputData):
    """ Prepare slices for features series.
    Example of result for time series [0, 1, 2, 3]:
    [[0, nan, nan, nan],
     [0,   1, nan, nan],
     [0,   1,   2, nan],
     [0,   1,   2,   3]]
    """
    nan_mask = np.triu(np.ones_like(input_data.features, dtype=bool), k=1)
    final_matrix = np.tril(input_data.features, k=0)
    final_matrix = np.array(final_matrix, dtype=float)
    final_matrix[nan_mask] = np.nan

    return final_matrix


def fix_elements_number(elements_to_take: int):
    if elements_to_take < 2:
        return 2
    return elements_to_take



import os

import numpy as np

from examples.advanced.sensitivity_analysis.dataset_access import get_scoring_data
from examples.advanced.sensitivity_analysis.pipelines_access import get_three_depth_manual_class_pipeline
from fedot.core.pipelines.pipeline import Pipeline
from fedot.sensitivity.node_sa_approaches import NodeDeletionAnalyze, NodeReplaceOperationAnalyze
from fedot.sensitivity.nodes_sensitivity import NodesAnalysis


def create_correct_path(path: str, dirname_flag: bool = False):
    """
    Create path with time which was created during the testing process.
    """

    for dirname in next(os.walk(os.path.curdir))[1]:
        if dirname.endswith(path):
            if dirname_flag:
                return dirname
            else:
                file = os.path.join(dirname, path + '.json')
                return file
    return None


def run_import_export_example(pipeline_path):
    # Prepare data to train the model
    train_data, test_data = get_scoring_data()

    # Get pipeline and fit it
    pipeline = get_three_depth_manual_class_pipeline()
    pipeline.fit_from_scratch(train_data)

    predicted_output = pipeline.predict(test_data)
    prediction_before_export = np.array(predicted_output.predict)
    print(f'Before export {prediction_before_export[:4]}')

    NodesAnalysis(pipeline, train_data, test_data,
                  approaches=[NodeDeletionAnalyze,
                              NodeReplaceOperationAnalyze]).analyze()

    # Export it
    pipeline.save(path=pipeline_path)

    # Import pipeline
    json_path_load = create_correct_path(pipeline_path)
    new_pipeline = Pipeline.from_serialized(json_path_load)

    predicted_output_after_export = new_pipeline.predict(test_data)
    prediction_after_export = np.array(predicted_output_after_export.predict)

    print(f'After import {prediction_after_export[:4]}')


if __name__ == '__main__':
    run_import_export_example(pipeline_path='import_export_sa')



from typing import List, Dict

from fedot.core.data.data import OutputData
from fedot.core.data.supplementary_data import SupplementaryData
from fedot.core.log import default_log
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.preprocessing.data_types import TableTypesCorrector


class SupplementaryDataMerger:
    def __init__(self, outputs: List['OutputData'], main_output: 'OutputData'):
        self.outputs = outputs
        self.main_output = main_output
        self.log = default_log(self)

    def merge(self) -> SupplementaryData:
        return SupplementaryData(
            is_main_target=self.main_output.supplementary_data.is_main_target,
            data_flow_length=self.calculate_dataflow_len(),
            features_mask=self.prepare_parent_mask(),
            previous_operations=None,  # is set by Node after merge
            was_preprocessed=self.all_preprocessed(),
            non_int_idx=None,  # is set elsewhere (by preprocessor or during pipeline fit/predict)
            column_types=self.merge_column_types()
        )

    def calculate_dataflow_len(self) -> int:
        """ Number of visited nodes is the max number among outputs plus 1 (the next operation). """
        return 1 + max(output.supplementary_data.data_flow_length for output in self.outputs)

    def all_preprocessed(self) -> bool:
        return all(output.supplementary_data.was_preprocessed for output in self.outputs)

    def prepare_parent_mask(self) -> Dict:
        """ The method for OutputData from multiple parent nodes prepares a field
        with encoded values. This allows distinguishing the source ancestor to
        which the data was attached.

        For example, a mask for two ancestors, each giving predictions
        in the form of tabular data with two columns will look like this:
        {'input_ids': [0, 0, 1, 1], 'flow_lens': [2, 2, 1, 1]}

        Similarly, for images the mask will specify where the channels came from.
        Mask for two images (1st with 2 channels and the 2nd with 3) will look like:
        {'input_ids': [0, 0, 1, 1, 1], 'flow_lens': [1, 1, 3, 3, 3]}

        :return features_mask: dict with mask for features/channels.
        A composite ID of two lists is used:
            - index of parent operation
            - dataflow length (number of nodes that data visited)
        """
        input_ids = []
        flow_lens = []
        for input_id, output in enumerate(self.outputs):
            # Get number of features (for tables and series) or channels (for images)
            table_shape = output.predict.shape
            if len(table_shape) > 1:
                num_features = table_shape[-1]  # we need last dimension
            else:
                num_features = 1

            # Each index specifies the source of the feature
            id_mask = [input_id] * num_features
            input_ids.extend(id_mask)

            # Keep dataflow length for each feature separately
            flow_mask = [output.supplementary_data.data_flow_length] * num_features
            flow_lens.extend(flow_mask)

        features_mask = {'input_ids': input_ids, 'flow_lens': flow_lens}
        return features_mask

    def merge_column_types(self) -> Dict:
        """ Store information about column types in tabular data for merged data """
        if self.main_output.data_type is not DataTypesEnum.table:
            # Data is not tabular
            return self.main_output.supplementary_data.column_types

        # Concatenate types for features columns and
        #  choose target type of the main target as the new target type
        new_features_types = []
        new_target_types = None
        for output in self.outputs:
            if output.supplementary_data.column_types is None:
                self.log.debug(f'Perform determination of column types in DataMerger')
                table_corr = TableTypesCorrector()
                output.supplementary_data.column_types = table_corr.prepare_column_types_info(output.predict,
                                                                                              output.target,
                                                                                              output.task)
            col_types = output.supplementary_data.column_types['features']
            new_features_types.extend(col_types)

            if output.supplementary_data.is_main_target:
                # Target can be None for predict stage
                new_target_types = output.supplementary_data.column_types.get('target')

        column_types = {'features': new_features_types, 'target': new_target_types}
        return column_types



from abc import ABC, abstractmethod
from typing import List, Optional, Union

from fedot.core.data.data import InputData
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.log import default_log
from fedot.core.optimisers.composer_requirements import ComposerRequirements
from fedot.core.optimisers.opt_history_objects.opt_history import OptHistory
from fedot.core.optimisers.optimizer import GraphOptimizer
from fedot.core.pipelines.pipeline import Pipeline


class Composer(ABC):
    """
    Base class used for receiving composite operations via optimization

    Args:
        optimizer: optimizer generated in :class:`~fedot.core.composer.ComposerBuilder`
        composer_requirements: requirements for composition process
    """

    def __init__(self, optimizer: GraphOptimizer, composer_requirements: Optional[ComposerRequirements] = None):
        self.composer_requirements = composer_requirements
        self.optimizer = optimizer
        self.log = default_log(self)

    @property
    def history(self) -> OptHistory:
        return self.optimizer.history

    @abstractmethod
    def compose_pipeline(self, data: Union[InputData, MultiModalData]) -> Union[Pipeline, List[Pipeline]]:
        """
        Run composition process for optimal pipeline structure search

        Args:
            data: Data used for problem solving

        Returns:
            For ``single-objective optimization`` -> the best pipeline.\n
            For ``multi-objective optimization`` -> a list of the best pipelines is returned.

        Notes:
            Returned pipelines are ordered by the descending primary metric (the first is the best).
        """

        raise NotImplementedError()



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict, Iterable, Optional, Union

import numpy as np
import seaborn as sns

from fedot.core.optimisers.opt_history_objects.opt_history import OptHistory
from fedot.core.repository.operation_types_repository import OperationTypesRepository, get_visualization_tags_map
from fedot.core.visualisation.graph_viz import GraphVisualizer
from fedot.core.visualisation.opt_history.utils import LabelsColorMapType
from fedot.core.visualisation.opt_viz import OptHistoryVisualizer


if TYPE_CHECKING:
    from fedot.core.dag.graph import Graph
    from fedot.core.optimisers.graph import OptGraph

    GraphType = Union[Graph, OptGraph]


def get_palette_based_on_default_tags() -> LabelsColorMapType:
    default_tags = [*OperationTypesRepository.DEFAULT_MODEL_TAGS, *OperationTypesRepository.DEFAULT_DATA_OPERATION_TAGS]
    p_1 = sns.color_palette('tab20')
    colour_period = 2  # diverge similar nearby colors
    p_1 = [p_1[i // (len(p_1) // colour_period) + i * colour_period % len(p_1)] for i in range(len(p_1))]
    p_2 = sns.color_palette('Set3')
    palette = np.vstack([p_1, p_2])
    palette_map = {tag: palette[i] for i, tag in enumerate(default_tags)}
    palette_map.update({None: 'mediumaquamarine'})
    return palette_map


def get_colors_by_tags(labels: Iterable[str]) -> LabelsColorMapType:
    tags_map = get_visualization_tags_map()
    new_map = {}
    for tag, operations in tags_map.items():
        new_map.update({operation: tag for operation in operations})
    tags_map = new_map

    palette = get_palette_based_on_default_tags()
    return {label: palette[tags_map.get(label)] for label in labels}


def get_pipeline_show_default_params() -> Dict[str, Any]:
    return {
        'node_color': get_colors_by_tags
    }


class PipelineHistoryVisualizer(OptHistoryVisualizer):
    def __init__(self, history: OptHistory, visuals_params: Optional[Dict[str, Any]] = None):
        visuals_params = visuals_params if visuals_params is not None else {}
        visuals_params['tags_map'] = visuals_params.get('tags_map') or get_visualization_tags_map()
        visuals_params['palette'] = visuals_params.get('palette') or get_palette_based_on_default_tags()
        visuals_params['graph_show_params'] = (visuals_params.get('graph_show_params') or
                                               get_pipeline_show_default_params())

        super().__init__(history, visuals_params)


class PipelineVisualizer(GraphVisualizer):
    def __init__(self, graph: GraphType, visuals_params: Optional[Dict[str, Any]] = None):
        visuals_params = visuals_params or {}
        default_visuals_params = get_pipeline_show_default_params()
        default_visuals_params.update(visuals_params)
        super().__init__(graph, default_visuals_params)



from typing import Iterator, Optional, Tuple, Type

import numpy as np
from sklearn.model_selection import KFold, TimeSeriesSplit
from sklearn.model_selection._split import _BaseKFold

from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.log import LoggerAdapter, default_log
from fedot.core.repository.dataset_types import DataTypesEnum


class OneFoldInputDataSplit:
    """ Perform one fold split (hold out) for InputData structures """

    def __init__(self):
        pass

    @staticmethod
    def input_split(input_data: InputData, **kwargs):
        # Train test split
        train_input, test_input = train_test_data_setup(input_data, **kwargs)

        yield train_input, test_input


class TsInputDataSplit(TimeSeriesSplit):
    """ Perform time series splitting for cross validation on InputData structures.
    The difference between TimeSeriesSplit (sklearn) and TsInputDataSplit can be
    demonstrated by an example:
    The time series [1, 2, 3, 4, 5, 6, 7, 8, 9, 10] must be splitted into 3
    parts, where the size of each fold for validation will be 2 elements.
    TimeSeriesSplit (return indices)
        train - [0, 1, 2, 3] test - [4, 5]
        train - [0, 1, 2, 3, 4, 5] test - [6, 7]
        train - [0, 1, 2, 3, 4, 5, 6, 7] test - [8, 9]
    TsInputDataSplit (return values of time series)
        train - [1, 2, 3, 4] test - [1, 2, 3, 4, 5, 6]
        train - [1, 2, 3, 4, 5, 6] test - [1, 2, 3, 4, 5, 6, 7, 8]
        train - [1, 2, 3, 4, 5, 6, 7, 8] test - [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
    """

    def __init__(self, validation_blocks: int, **params):
        super().__init__(**params)
        self.validation_blocks = validation_blocks
        self.params = params

    def input_split(self, input_data: InputData) -> Iterator[Tuple[InputData, InputData]]:
        """ Splitting into datasets for train and validation using
        "in-sample forecasting" algorithm

        :param input_data: InputData for splitting
        """
        # Transform InputData into numpy array
        data_for_split = np.array(input_data.target)

        for train_ids, test_ids in super().split(data_for_split):
            if len(train_ids) <= len(test_ids):
                raise ValueError("Train size will be too small with selected number of folds and validation blocks")
            # Return train part by ids
            train_features, train_target = _ts_data_by_index(train_ids, train_ids, input_data)
            train_data = InputData(idx=np.arange(0, len(train_target)),
                                   features=train_features, target=train_target,
                                   task=input_data.task,
                                   data_type=input_data.data_type,
                                   supplementary_data=input_data.supplementary_data)

            # Unit all ids for "in-sample validation"
            all_ids = np.hstack((train_ids, test_ids))
            # In-sample validation dataset
            val_features, val_target = _ts_data_by_index(all_ids, all_ids, input_data)
            validation_data = InputData(idx=np.arange(0, len(val_target)),
                                        features=val_features, target=val_target,
                                        task=input_data.task,
                                        data_type=input_data.data_type,
                                        supplementary_data=input_data.supplementary_data)

            yield train_data, validation_data


def tabular_cv_generator(data: InputData,
                         folds: int,
                         split_method: Type[_BaseKFold] = KFold) -> Iterator[Tuple[InputData, InputData]]:
    """ The function for splitting data into a train and test samples
        in the InputData format for KFolds cross validation. The function
        return a generator of tuples, consisting of a pair of train, test.

    :param data: InputData for train and test splitting
    :param folds: number of folds
    :param split_method: method to split data (f.e. stratify KFold)

    :return Iterator[InputData, InputData]: return split train/test data
    """
    kf = split_method(n_splits=folds, shuffle=True, random_state=42)

    for train_idxs, test_idxs in kf.split(data.features, data.target):
        train_features, train_target = _table_data_by_index(train_idxs, data)
        test_features, test_target = _table_data_by_index(test_idxs, data)

        idx_for_train = np.arange(0, len(train_features))
        idx_for_test = np.arange(0, len(test_features))

        train_data = InputData(idx=idx_for_train,
                               features=train_features,
                               target=train_target,
                               task=data.task,
                               data_type=data.data_type,
                               supplementary_data=data.supplementary_data)
        test_data = InputData(idx=idx_for_test,
                              features=test_features,
                              target=test_target,
                              task=data.task,
                              data_type=data.data_type,
                              supplementary_data=data.supplementary_data)

        yield train_data, test_data


def ts_cv_generator(data: InputData, folds: int,
                    validation_blocks: int = 1, log: Optional[LoggerAdapter] = None) \
        -> Iterator[Tuple[InputData, InputData]]:
    """ Splitting data for time series cross validation

    :param data: source InputData with time series data type
    :param folds: number of folds
    :param validation_blocks: number of validation block per each fold
    :param log: log object
    """
    if not log:
        log = default_log(prefix=__name__)
    validation_blocks = int(validation_blocks)
    # Forecast horizon for each fold
    horizon = data.task.task_params.forecast_length * validation_blocks

    try:
        tscv = TsInputDataSplit(gap=0, validation_blocks=validation_blocks,
                                n_splits=folds, test_size=horizon)
        for train_data, test_data in tscv.input_split(data):
            yield train_data, test_data
    except ValueError:
        log.info(f'Time series length too small for cross validation with {folds} folds. Perform one fold validation')
        # Perform one fold validation (folds parameter will be ignored)

        one_fold_split = OneFoldInputDataSplit()
        for train_data, test_data in one_fold_split.input_split(data, validation_blocks=validation_blocks):
            yield train_data, test_data


def _table_data_by_index(index, values: InputData):
    """ Allow to get tabular data by indexes of elements """
    features = values.features[index, :]
    target = np.take(values.target, index)

    return features, target


def _ts_data_by_index(train_ids, test_ids, data):
    """ Allow to get time series data by indexes of elements """
    features = data.features[train_ids]
    target = data.target[test_ids]

    # Use only the first time-series as target for multi_ts
    if data.data_type == DataTypesEnum.multi_ts:
        target = target[:, 0]

    return features, target



from typing import Optional

from fedot.core.data.data import OutputData
from fedot.core.operations.operation import Operation
from fedot.core.repository.operation_types_repository import OperationMetaInfo, OperationTypesRepository


class DataOperation(Operation):
    """Class with ``fit``/``predict`` methods defining the evaluation strategy for the task

    Args:
        operation_type: name of the data operation
    """
    def __init__(self, operation_type: str):
        super().__init__(operation_type)
        self.operations_repo = OperationTypesRepository('data_operation')

    @property
    def metadata(self) -> OperationMetaInfo:
        operation_info = self.operations_repo.operation_info_by_id(self.operation_type)
        if not operation_info:
            raise ValueError(f'Data operation {self.operation_type} not found')
        return operation_info

    @staticmethod
    def assign_tabular_column_types(output_data: OutputData, output_mode: str) -> OutputData:
        """Assign new column types if it necessary.
        By default, all data operations must define column types at lower levels (:obj:`EvalStrategies` and :obj:`Implementations`).
        In some cases the previously defined data types are passed.
        """
        return output_data



import os
import timeit
import warnings

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from sklearn.metrics import mean_absolute_error

from examples.simple.time_series_forecasting.ts_pipelines import ts_complex_dtreg_pipeline
from fedot.core.data.data import InputData
from fedot.core.pipelines.tuning.tuner_builder import TunerBuilder
from fedot.core.pipelines.tuning.unified import PipelineTuner
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.quality_metrics_repository import RegressionMetricsEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum, TsForecastingParams
from fedot.core.utils import fedot_project_root

warnings.filterwarnings('ignore')


def prepare_train_test_input(train_part, len_forecast):
    """ Function return prepared data for fit and predict

    :param len_forecast: forecast length
    :param train_part: time series which can be used as predictors for train

    :return train_input: Input Data for fit
    :return predict_input: Input Data for predict
    :return task: Time series forecasting task with parameters
    """

    task = Task(TaskTypesEnum.ts_forecasting,
                TsForecastingParams(forecast_length=len_forecast))

    train_input = InputData(idx=np.arange(0, len(train_part)),
                            features=train_part,
                            target=train_part,
                            task=task,
                            data_type=DataTypesEnum.ts)

    start_forecast = len(train_part)
    end_forecast = start_forecast + len_forecast
    predict_input = InputData(idx=np.arange(start_forecast, end_forecast),
                              features=train_part,
                              target=None,
                              task=task,
                              data_type=DataTypesEnum.ts)

    return train_input, predict_input, task


def run_tuning_test(pipeline, train_input, predict_input, test_data, task, show_metrics=True):
    """
    Function for predicting values in a time series

    :param pipeline: TsForecastingPipeline object
    :param train_input: InputData for fit
    :param predict_input: InputData for predict
    :param test_data: numpy array for validation
    :param task: Ts_forecasting task
    :param show_metrics: is there need to print metrics before and after tuning

    :return amount_of_seconds time spent on tuning
    :return mae_before MAE metric of pipeline without tuning
    :return mae_after MAE metric of pipeline with tuning
    """

    pipeline.fit_from_scratch(train_input)

    # Predict
    predicted_values = pipeline.predict(predict_input)
    old_predicted_values = predicted_values.predict
    cv_folds = 3
    validation_blocks = 2

    start_time = timeit.default_timer()
    pipeline_tuner = TunerBuilder(task)\
        .with_tuner(PipelineTuner)\
        .with_metric(RegressionMetricsEnum.MAE)\
        .with_cv_folds(cv_folds) \
        .with_validation_blocks(validation_blocks)\
        .with_iterations(20) \
        .build(train_input)
    pipeline = pipeline_tuner.tune(pipeline)
    print(pipeline.print_structure())
    amount_of_seconds = timeit.default_timer() - start_time
    print(f'\nTuning pipline on {amount_of_seconds:.2f} seconds\n')

    # Fit pipeline on the entire train data
    pipeline.fit_from_scratch(train_input)
    # Predict
    predicted_values = pipeline.predict(predict_input)
    new_predicted_values = predicted_values.predict
    new_predicted_values = np.ravel(np.array(new_predicted_values))
    old_predicted_values = np.ravel(np.array(old_predicted_values))

    mae_before = mean_absolute_error(test_data, old_predicted_values)
    mae_after = mean_absolute_error(test_data, new_predicted_values)

    if show_metrics:
        print(f'MAE before tuning - {mae_before:.4f}')
        print(f'MAE after tuning - {mae_after:.4f}\n')

    return amount_of_seconds, mae_before, mae_after


def visualize(tuned, no_tuned, time, method_name):
    ind = np.arange(len(tuned))
    width = 0.4

    fig, ax1 = plt.subplots()

    color = 'tab:red'
    ax1.bar(ind, no_tuned, width, fc=(0, 0, 1, 0.5), label='No tuning')
    ax1.bar(ind + width, tuned, width, fc=(1, 0, 0, 0.5), label='Tuned')
    ax1.set_ylabel('MAE', color=color)
    ax1.tick_params(axis='y', labelcolor=color)

    ax2 = ax1.twinx()

    color = 'tab:green'
    ax2.plot(ind + width / 2, time, color=color)
    ax2.set_ylabel('time, s', color=color)
    ax2.tick_params(axis='y', labelcolor=color)

    plt.title(f'MAE with tuning process based on {method_name}')
    fig.legend()
    fig.tight_layout()
    plt.show()


def run_tuning_comparison(n_repits=10, ts_size=1000, forecast_length=50, visualization=True):
    file_path = os.path.join(str(fedot_project_root()), 'cases/data/time_series/temperature.csv')
    df = pd.read_csv(file_path)
    time_series = np.array(df['value'])[:ts_size]

    # Train/test split
    train_part = time_series[:-forecast_length]
    test_part = time_series[-forecast_length:]

    # Prepare data for train and test
    train_input, predict_input, task = prepare_train_test_input(train_part, forecast_length)

    nodes_names = ['sparse_lagged', 'lagged']
    for name in nodes_names:

        # set lists for data collecting
        time_list = []
        mae_no_tuning = []
        mae_tuning = []

        test_part = np.ravel(test_part)
        # tuning calculations for averaging
        for i in range(n_repits):
            pipeline = ts_complex_dtreg_pipeline(name)
            amount_of_seconds, mae_before, mae_after = run_tuning_test(pipeline,
                                                                       train_input,
                                                                       predict_input,
                                                                       test_part,
                                                                       task)
            time_list.append(amount_of_seconds)
            mae_no_tuning.append(mae_before)
            mae_tuning.append(mae_after)

        if visualization:
            visualize(mae_tuning, mae_no_tuning, time_list, name)

        print(f'Mean time: {np.array(time_list).mean()}')
        print(f'Mean MAE without tuning: {np.array(mae_no_tuning).mean()}')
        print(f'Mean MAE with tuning: {np.array(mae_tuning).mean()}')


if __name__ == '__main__':
    # On large time series the speed of sparse_lagged increase (ts_size parameter)
    run_tuning_comparison(n_repits=10, ts_size=1000, forecast_length=50, visualization=True)



from abc import ABC, abstractmethod
from copy import deepcopy
from typing import Union, Optional

import numpy as np
import pandas as pd

from fedot.core.data.data import InputData, array_to_input_data
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum

FeaturesType = Union[str, np.ndarray, pd.DataFrame, InputData, dict, tuple]
TargetType = Union[str, np.ndarray, pd.Series, dict]


class DataDefiner:

    def __init__(self, strategy) -> None:
        self._strategy = strategy

    @property
    def strategy(self):
        return self._strategy

    @strategy.setter
    def strategy(self, strategy) -> None:
        self._strategy = strategy

    def define_data(self, features: FeaturesType,
                    ml_task: Task,
                    target: Optional[str] = None,
                    is_predict: bool = False) -> None:
        return self._strategy.define_data(features,
                                          ml_task,
                                          target,
                                          is_predict)


class StrategyDefineData(ABC):
    @abstractmethod
    def define_data(self, features: Union[tuple, str, np.ndarray, pd.DataFrame, InputData],
                    ml_task: Task,
                    target: str = None,
                    is_predict: bool = False):
        pass


class FedotStrategy(StrategyDefineData):
    def define_data(self, features: Union[InputData, MultiModalData],
                    ml_task: Task,
                    target: str = None,
                    is_predict: bool = False) -> Union[InputData, MultiModalData]:
        # InputData or MultiModalData format for input data
        data = deepcopy(features)
        data.task = ml_task
        return data


class TupleStrategy(StrategyDefineData):
    def define_data(self, features: tuple,
                    ml_task: Task,
                    target: str = None,
                    is_predict: bool = False) -> InputData:
        data = array_to_input_data(features_array=features[0],
                                   target_array=features[1],
                                   task=ml_task)
        return data


class PandasStrategy(StrategyDefineData):
    """ Class for wrapping pandas DataFrames into InputData """

    def define_data(self, features: pd.DataFrame,
                    ml_task: Task,
                    target: str = None,
                    is_predict: bool = False) -> InputData:
        if target is None:
            target = np.array([])

        if isinstance(target, str) and target in features.columns:
            target_array = features[target]
            features = features.drop(columns=[target])
        else:
            target_array = target

        data = array_to_input_data(features_array=np.asarray(features),
                                   target_array=np.asarray(target_array),
                                   task=ml_task)
        return data


class NumpyStrategy(StrategyDefineData):
    def define_data(self, features: np.ndarray,
                    ml_task: Task,
                    target: Optional[int] = None,
                    is_predict: bool = False) -> InputData:
        # numpy format for input data
        if target is None:
            target = np.array([])

        if isinstance(target, int):
            target_array = features[target]
            features = np.delete(features, target, axis=1)
        else:
            target_array = target

        data = array_to_input_data(features_array=features,
                                   target_array=target_array,
                                   task=ml_task)
        return data


class CsvStrategy(StrategyDefineData):
    def define_data(self, features: str,
                    ml_task: Task,
                    target: str = None,
                    is_predict: bool = False) -> InputData:
        # CSV files as input data, by default - table data

        data_type = DataTypesEnum.table
        if ml_task.task_type == TaskTypesEnum.ts_forecasting:
            # For time series forecasting format - time series
            data = InputData.from_csv_time_series(task=ml_task,
                                                  file_path=features,
                                                  target_column=target,
                                                  is_predict=is_predict)
        else:
            # Make default features table
            # CSV files as input data
            data = InputData.from_csv(features, task=ml_task,
                                      target_columns=target,
                                      data_type=data_type)
        return data


class MultimodalStrategy(StrategyDefineData):
    """
    Gets dict of NumPy arrays or InputData sources as input data
    and returns MultiModalData object with source names defined by data type
    """
    source_name_by_type = {'table': 'data_source_table',
                           'ts': 'data_source_ts',
                           'multi_ts': 'data_source_ts',
                           'text': 'data_source_text',
                           'image': 'data_source_image'}

    def define_data(self, features: dict,
                    ml_task: Task,
                    target: str = None,
                    is_predict: bool = False,
                    idx=None) -> MultiModalData:

        # change data type to InputData
        for source in features:
            if not isinstance(features[source], InputData):
                features[source] = array_to_input_data(features_array=features[source], target_array=target,
                                                       task=ml_task, idx=idx)
        # create labels for data sources
        sources = dict((f'{self.source_name_by_type.get(features[data_part_key].data_type.name)}/{data_part_key}',
                        data_part)
                       for (data_part_key, data_part) in features.items())
        data = MultiModalData(sources)
        return data


def data_strategy_selector(features, target, ml_task: Task = None, is_predict: bool = None):
    data_type = type(features)

    data = DataDefiner(_strategy_dispatch[data_type]())
    return data.define_data(features, ml_task, target, is_predict)


_strategy_dispatch = {InputData: FedotStrategy,
                      MultiModalData: FedotStrategy,
                      tuple: TupleStrategy,
                      pd.DataFrame: PandasStrategy,
                      np.ndarray: NumpyStrategy,
                      str: CsvStrategy,
                      dict: MultimodalStrategy}



from functools import partial
from inspect import signature
from typing import Callable, List, Sequence, Optional, Union

from fedot.core.adapter import BaseOptimizationAdapter
from fedot.core.dag.graph import Graph
from fedot.core.dag.graph_verifier import GraphVerifier, VerifierRuleType
from fedot.core.dag.verification_rules import (
    has_no_cycle,
    has_no_isolated_nodes,
    has_no_self_cycled_nodes,
    has_one_root
)
from fedot.core.pipelines.adapters import PipelineAdapter
from fedot.core.optimisers.graph import OptGraph
from fedot.core.pipelines.verification_rules import (
    has_correct_data_connections,
    has_correct_data_sources,
    has_correct_operations_for_task,
    has_final_operation_as_model,
    has_no_conflicts_after_class_decompose,
    has_no_conflicts_during_multitask,
    has_no_conflicts_in_decompose,
    has_no_conflicts_with_data_flow,
    has_no_data_flow_conflicts_in_ts_pipeline,
    has_primary_nodes,
    is_pipeline_contains_ts_operations,
    only_non_lagged_operations_are_primary
)
from fedot.core.repository.tasks import TaskTypesEnum

common_rules = [has_one_root,
                has_no_cycle,
                has_no_self_cycled_nodes,
                has_no_isolated_nodes,
                has_primary_nodes,
                has_correct_operations_for_task,
                has_final_operation_as_model,
                has_no_conflicts_with_data_flow,
                has_no_conflicts_in_decompose,
                has_correct_data_connections,
                has_correct_data_sources]

ts_rules = [is_pipeline_contains_ts_operations,
            only_non_lagged_operations_are_primary,
            has_no_data_flow_conflicts_in_ts_pipeline]

class_rules = [has_no_conflicts_during_multitask,
               has_no_conflicts_after_class_decompose]


def verifier_for_task(task_type: Optional[TaskTypesEnum] = None, adapter: Optional[BaseOptimizationAdapter] = None):
    adapter = adapter or PipelineAdapter()
    return GraphVerifier(rules_by_task(task_type), adapter)


def rules_by_task(task_type: Optional[TaskTypesEnum],
                  rules: Sequence[VerifierRuleType] = ()) -> Sequence[VerifierRuleType]:
    tmp_rules = []

    # provide additional args if necessary
    # somewhat hack-y, made for `has_correct_operations_for_task`
    for rule in (rules or common_rules):
        if 'task_type' in signature(rule).parameters:
            tmp_rules.append(partial(rule, task_type=task_type))
        else:
            tmp_rules.append(rule)

    if task_type is TaskTypesEnum.ts_forecasting:
        tmp_rules.extend(ts_rules)
    elif task_type is TaskTypesEnum.classification:
        tmp_rules.extend(class_rules)

    return tmp_rules


def verify_pipeline(graph: Union[Graph, OptGraph],
                    rules: List[Callable] = None,
                    task_type: Optional[TaskTypesEnum] = None,
                    raise_on_failure: bool = False):
    """Method for validation of graphs with default rules.
    NB: It is preserved for simplicity, use graph checker instead."""
    return GraphVerifier(rules_by_task(task_type, rules),
                         PipelineAdapter(),
                         raise_on_failure).verify(graph)



import warnings

import numpy as np
import pandas as pd
from sklearn.metrics import mean_absolute_error, mean_squared_error
from sklearn.model_selection import train_test_split

from fedot.core.data.data import InputData
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.tuning.tuner_builder import TunerBuilder
from fedot.core.pipelines.tuning.unified import PipelineTuner
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.quality_metrics_repository import RegressionMetricsEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum
from fedot.core.utils import fedot_project_root

warnings.filterwarnings('ignore')


def get_refinement_pipeline():
    """ Create five-level pipeline with decompose operation """
    node_encoding = PrimaryNode('one_hot_encoding')
    node_scaling = SecondaryNode('scaling', nodes_from=[node_encoding])
    node_lasso = SecondaryNode('lasso', nodes_from=[node_scaling])
    node_decompose = SecondaryNode('decompose', nodes_from=[node_scaling, node_lasso])
    node_dtreg = SecondaryNode('dtreg', nodes_from=[node_decompose])
    node_dtreg.parameters = {'max_depth': 3}
    final_node = SecondaryNode('ridge', nodes_from=[node_lasso, node_dtreg])

    pipeline = Pipeline(final_node)
    return pipeline


def get_non_refinement_pipeline():
    node_encoding = PrimaryNode('one_hot_encoding')
    node_scaling = SecondaryNode('scaling', nodes_from=[node_encoding])

    node_lasso = SecondaryNode('lasso', nodes_from=[node_scaling])
    node_dtreg = SecondaryNode('dtreg', nodes_from=[node_scaling])
    node_dtreg.parameters = {'max_depth': 3}

    final_node = SecondaryNode('ridge', nodes_from=[node_lasso, node_dtreg])

    pipeline = Pipeline(final_node)
    return pipeline


def prepare_input_data(features, target):
    """ Function create InputData with features """
    x_data_train, x_data_test, y_data_train, y_data_test = train_test_split(
        features,
        target,
        test_size=0.2,
        shuffle=True,
        random_state=10)
    y_data_test = np.ravel(y_data_test)

    # Define regression task
    task = Task(TaskTypesEnum.regression)

    # Prepare data to train the model
    train_input = InputData(idx=np.arange(0, len(x_data_train)),
                            features=x_data_train,
                            target=y_data_train,
                            task=task,
                            data_type=DataTypesEnum.table)

    predict_input = InputData(idx=np.arange(0, len(x_data_test)),
                              features=x_data_test,
                              target=y_data_test,
                              task=task,
                              data_type=DataTypesEnum.table)

    return train_input, predict_input, task


def run_river_experiment(file_path, with_tuning=False):
    """ Function launch example with experimental features of the FEDOT framework

    :param file_path: path to the csv file
    :param with_tuning: is it need to tune pipelines or not
    """

    # Read dataframe and prepare train and test data
    df = pd.read_csv(file_path)
    features = np.array(df[['level_station_1', 'mean_temp', 'month', 'precip']])
    target = np.array(df['level_station_2']).reshape((-1, 1))

    # Prepare InputData for train and test
    train_input, predict_input, task = prepare_input_data(features, target)
    y_data_test = predict_input.target

    # Get refinement pipeline
    r_pipeline = get_refinement_pipeline()
    non_pipeline = get_non_refinement_pipeline()

    if with_tuning:
        tuner = TunerBuilder(task)\
            .with_tuner(PipelineTuner)\
            .with_metric(RegressionMetricsEnum.MAE)\
            .with_iterations(100)\
            .build(train_input)
        r_pipeline = tuner.tune(r_pipeline)
        non_pipeline = tuner.tune(non_pipeline)

    # Fit it
    r_pipeline.fit(train_input)
    non_pipeline.fit(train_input)

    # Predict
    predicted_values = r_pipeline.predict(predict_input)
    r_preds = predicted_values.predict

    # Predict
    predicted_values = non_pipeline.predict(predict_input)
    non_preds = predicted_values.predict

    y_data_test = np.ravel(y_data_test)

    mse_value = mean_squared_error(y_data_test, r_preds, squared=False)
    mae_value = mean_absolute_error(y_data_test, r_preds)
    print(f'RMSE with decompose - {mse_value:.2f}')
    print(f'MAE with decompose - {mae_value:.2f}\n')

    mse_value_non = mean_squared_error(y_data_test, non_preds, squared=False)
    mae_value_non = mean_absolute_error(y_data_test, non_preds)
    print(f'RMSE non decompose - {mse_value_non:.2f}')
    print(f'MAE non decompose - {mae_value_non:.2f}\n')


if __name__ == '__main__':
    run_river_experiment(file_path=f'{fedot_project_root()}/cases/data/river_levels/station_levels.csv',
                         with_tuning=True)



from fedot.core.data.data import data_type_is_ts, data_type_is_multi_ts, data_type_is_image


def exclude_ts(preprocessing_function):
    """ Decorator for time series type checking """
    def wrapper(self, input_data, source_name, *args, **kwargs):
        if data_type_is_ts(input_data):
            return input_data
        return preprocessing_function(self, input_data, source_name, *args, **kwargs)
    return wrapper


def exclude_multi_ts(preprocessing_function):
    """ Decorator for multi time series (not to be confused with multivariate ts) type checking """
    def wrapper(self, input_data, source_name, *args, **kwargs):
        if data_type_is_multi_ts(input_data):
            return input_data
        return preprocessing_function(self, input_data, source_name, *args, **kwargs)
    return wrapper


def exclude_image(preprocessing_function):
    """ Decorator for image type checking """
    def wrapper(self, input_data, source_name, *args, **kwargs):
        if data_type_is_image(input_data):
            return input_data
        return preprocessing_function(self, input_data, source_name, *args, **kwargs)
    return wrapper




import warnings
from typing import Optional
from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.evaluation_interfaces import EvaluationStrategy
from fedot.core.operations.evaluation.operation_implementations.models.custom_model import CustomModelImplementation
from fedot.core.operations.operation_parameters import OperationParameters

warnings.filterwarnings("ignore", category=UserWarning)


class CustomModelStrategy(EvaluationStrategy):
    """
    This class defines the default model container for custom of domain-specific implementations
    :param str operation_type: rudimentary of parent - type of the operation defined in operation or
           data operation repositories
    :param dict params: hyperparameters to fit the model with
    """

    def __init__(self, operation_type: Optional[str], params: Optional[OperationParameters] = None):
        super().__init__(operation_type, params)
        self.operation_impl = CustomModelImplementation(params)

    def fit(self, train_data: InputData):
        """ Fit method for custom strategy"""
        self.operation_impl.fit(train_data)
        return self.operation_impl

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:

        prediction = trained_operation.predict(predict_data)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def predict_for_fit(self, trained_operation, predict_data: InputData) -> OutputData:

        prediction = trained_operation.predict_for_fit(predict_data)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def _convert_to_operation(self, operation_type: str):
        return CustomModelImplementation



import datetime
import os
import random

import pandas as pd

from fedot.utilities.requirements_notificator import warn_requirement

try:
    import openpyxl
except ImportError:
    warn_requirement('openpyxl', 'fedot[examples]', should_raise=True)

from datetime import timedelta

import numpy as np

from fedot.core.composer.composer_builder import ComposerBuilder
from fedot.core.optimisers.gp_comp.pipeline_composer_requirements import PipelineComposerRequirements
from fedot.core.data.data import InputData
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.repository.operation_types_repository import OperationTypesRepository
from fedot.core.repository.quality_metrics_repository import ClassificationMetricsEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum
from fedot.core.utils import (
    ensure_directory_exists,
    fedot_project_root,
    get_split_data_paths,
    probs_to_labels,
    save_file_to_csv,
    split_data
)
from sklearn.metrics import roc_auc_score as roc_auc

random.seed(1)
np.random.seed(1)


def create_multi_clf_examples_from_excel(file_path: str, return_df: bool = False):
    """ Return dataframe from excel file or path to the csv file """
    df = pd.read_excel(file_path, engine='openpyxl')
    train, test = split_data(df)
    file_dir_name = file_path.replace('.', '/').split('/')[-2]
    file_csv_name = f'{file_dir_name}.csv'
    directory_names = ['examples', 'data', file_dir_name]

    # Check does obtained directory exist or not
    ensure_directory_exists(directory_names)
    if return_df:
        # Need to return dataframe and path to the file in csv format
        path = os.path.join(directory_names[0], directory_names[1], directory_names[2], file_csv_name)
        full_file_path = os.path.join(str(fedot_project_root()), path)
        save_file_to_csv(df, full_file_path)
        return df, full_file_path
    else:
        # Need to return only paths to the files with train and test data
        full_train_file_path, full_test_file_path = get_split_data_paths(directory_names)
        save_file_to_csv(train, full_train_file_path)
        save_file_to_csv(train, full_test_file_path)
        return full_train_file_path, full_test_file_path


def get_model(train_file_path: str, cur_lead_time: datetime.timedelta = timedelta(seconds=60)):
    task = Task(task_type=TaskTypesEnum.classification)
    dataset_to_compose = InputData.from_csv(train_file_path, task=task)

    # the search of the models provided by the framework
    # that can be used as nodes in a pipeline for the selected task
    models_repo = OperationTypesRepository()
    available_model_types = models_repo.suitable_operation(task_type=task.task_type, tags=['simple'])

    metric_function = ClassificationMetricsEnum.ROCAUC_penalty

    composer_requirements = PipelineComposerRequirements(
        primary=available_model_types, secondary=available_model_types,
        timeout=cur_lead_time)

    # Create the genetic programming-based composer, that allow to find
    # the optimal structure of the composite model
    builder = ComposerBuilder(task).with_requirements(composer_requirements).with_metrics(metric_function)
    composer = builder.build()

    # run the search of best suitable model
    pipeline_evo_composed = composer.compose_pipeline(data=dataset_to_compose)
    pipeline_evo_composed.fit(input_data=dataset_to_compose)

    return pipeline_evo_composed


def apply_model_to_data(model: Pipeline, data_path: str):
    df, file_path = create_multi_clf_examples_from_excel(data_path, return_df=True)
    dataset_to_apply = InputData.from_csv(file_path, target_columns=None)
    evo_predicted = model.predict(dataset_to_apply)
    df['forecast'] = probs_to_labels(evo_predicted.predict)
    return df


def validate_model_quality(model: Pipeline, data_path: str):
    dataset_to_validate = InputData.from_csv(data_path)
    predicted_labels = model.predict(dataset_to_validate).predict

    roc_auc_valid = round(roc_auc(y_true=test_data.target,
                                  y_score=predicted_labels,
                                  multi_class='ovo',
                                  average='macro'), 3)
    return roc_auc_valid


if __name__ == '__main__':
    file_path_first = r'../../data/example1.xlsx'
    file_path_second = r'../../data/example2.xlsx'
    file_path_third = r'../../data/example3.xlsx'

    train_file_path, test_file_path = create_multi_clf_examples_from_excel(file_path_first)
    test_data = InputData.from_csv(test_file_path)

    fitted_model = get_model(train_file_path)

    fitted_model.show()

    roc_auc = validate_model_quality(fitted_model, test_file_path)
    print(f'ROC AUC metric is {roc_auc}')

    final_prediction_first = apply_model_to_data(fitted_model, file_path_second)
    print(final_prediction_first['forecast'])

    final_prediction_second = apply_model_to_data(fitted_model, file_path_third)
    print(final_prediction_second['forecast'])



from typing import Callable, Optional

from fedot.core.operations.evaluation. \
    operation_implementations.implementation_interfaces import ModelImplementation
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.repository.dataset_types import DataTypesEnum


class CustomModelImplementation(ModelImplementation):
    """
    Implementation of container for custom model, which is presented as function with
    input train_data(np.array), test_data(np.array), parameters(dict)
    output type specification DataTypesEnum (string - 'ts', 'table', 'image', 'text')
    into parameters dictionary {'model_predict': function, 'model_fit': function}
    """

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        if not self.params:
            raise ValueError('There is no specified parameters for custom model!')
        self.fitted_model = None

    @property
    def model_predict(self) -> Callable:
        if 'model_predict' in self.params.keys():
            model_predict = self.params.get('model_predict')
            if not isinstance(model_predict, Callable):
                self.log.warning('Input model_predict is not Callable')
            return model_predict
        else:
            raise ValueError('There is no key word "model_predict" for model definition in input dictionary.')

    @property
    def model_fit(self) -> Optional[Callable]:
        # custom model can be without fitting
        model_fit = None
        if 'model_fit' in self.params.keys():
            model_fit = self.params.get('model_fit')
            if not isinstance(model_fit, Callable):
                raise ValueError('Input model is not Callable')
        return model_fit

    def fit(self, input_data):
        """ Fit method for custom model implementation """
        if self.model_fit:
            self.fitted_model = self.model_fit(input_data.idx, input_data.features, input_data.target, self.params)
        return self.fitted_model

    def predict(self, input_data):
        try:
            predict, output_type = self.model_predict(self.fitted_model,
                                                      input_data.idx,
                                                      input_data.features,
                                                      self.params)
            if (input_data.data_type == DataTypesEnum.ts and
                    input_data.target is not None and len(input_data.target.shape) > 1):
                # change target after custom model is processed
                input_data.target = input_data.target[:, 0]
            output_type = DataTypesEnum[output_type]
        except Exception as e:
            raise TypeError(f'{e}\nInput model has incorrect behaviour. Check type hints for model: \
                                    Callable[[any, np.array, dict], np.array, str]')
        output_data = self._convert_to_output(input_data,
                                              predict=predict,
                                              data_type=output_type)
        return output_data

    def predict_for_fit(self, input_data):
        output_type = input_data.data_type
        # if there is no need in fitting custom model and it is fit call
        if not self.model_fit:
            predict = input_data.features
            output_data = self._convert_to_output(input_data,
                                                  predict=predict,
                                                  data_type=output_type)
            return output_data
        else:
            return self.predict(input_data)



import argparse
from argparse import RawTextHelpFormatter
from fedot.core.repository.tasks import TsForecastingParams
from fedot.api.main import Fedot
import os


def create_parser(arguments_list):
    """ Function for creating parser object and adding arguments"""
    parser = argparse.ArgumentParser(description="__________________ FEDOT API console call __________________",
                                     formatter_class=RawTextHelpFormatter)

    for argument_dict in arguments_list:
        add_args_to_parser(parser=parser, **argument_dict)
    return parser


def add_args_to_parser(parser, tag: str, help: str, required: bool = False, is_list: bool = False):
    """ Function for formalized parser arguments creation
    :param parser: ArgumentParser class object
    :param tag: name of parameter to parse
    :param help: help text for parameter
    :param required: is the parameter required or not
    :param is_list: does the parameter contain multiple input
    """
    if is_list is True:
        nargs = '*'
    else:
        nargs = None
    parser.add_argument(tag, help=help, required=required, nargs=nargs)


def separate_argparse_to_fedot(parameters):
    """ Function for separating argparse parameters on fedot fit/predict/composer"""
    main_params = {}
    fit_params = {}
    for arg in vars(parameters):
        if arg in main_params_names:
            main_params[keys_names[arg]] = getattr(parameters, arg)
        if arg in fit_params_names:
            fit_params[keys_names[arg]] = getattr(parameters, arg)

    if main_params['problem'] == 'ts_forecasting' and getattr(parameters, 'for_len') is not None:
        main_params['task_params'] = TsForecastingParams(forecast_length=int(getattr(parameters, 'for_len')))
    elif main_params['problem'] == 'ts_forecasting' and getattr(parameters, 'for_len') is None:
        raise ValueError("Forecast length (for_len) is necessary parameter for ts_forecasting problem")

    if main_params['with_tuning'] == '1':
        main_params['with_tuning'] = True
    else:
        main_params['with_tuning'] = False
    return main_params, fit_params


def preprocess_keys(parameters: dict):
    """ Function that removes nan parameters from input
    and converts numbers from strings
    :param parameters: argparse filtered keys
    """
    for k, v in list(parameters.items()):
        if v is None:
            del parameters[k]
        elif type(v) is not bool:
            try:
                parameters[k] = float(v)
            except Exception:
                pass


def run_fedot(parameters, main_params, fit_params, save_predictions=True):
    """ Function for run prediction on fedot """
    model = Fedot(**main_params)
    print("\nFitting start...")
    model.fit(**fit_params)
    print("\nPrediction start...")
    prediction = model.predict(features=getattr(parameters, 'test'), in_sample=False, save_predictions=save_predictions)
    print(f"\nPrediction saved at {os.getcwd()}\\predictions.csv")
    return prediction


# parameters to init Fedot class
main_params_names = ['problem', 'timeout', 'seed', 'depth', 'arity', 'popsize', 'gen_num',
                     'opers', 'tuning', 'cv_folds', 'val_bl', 'hist_path', 'preset']
# parameters to fit model
fit_params_names = ['train', 'target']

# dictionary with keys for parser creation
arguments_dicts = [{'tag': '--problem',
                    'help': 'The name of modelling problem to solve: \n'
                            'classification;\n'
                            'regression;\n'
                            'ts_forecasting;\n'
                            'clustering',
                    'required': True},
                   {'tag': '--train',
                    'help': 'Path to train data file',
                    'required': True},
                   {'tag': '--test',
                    'help': 'Path to test data file',
                    'required': True},
                   {'tag': '--preset',
                    'help': 'Name of preset for model building: \n'
                            'light;\n'
                            'light_steady_state;\n'
                            'ultra_light;\n'
                            'ultra_steady_state;\n'
                            'ts;\n'
                            'gpu'},
                   {'tag': '--timeout',
                    'help': 'Time for model design (in minutes)'},
                   {'tag': '--seed',
                    'help': 'Value for fixed random seed'},
                   {'tag': '--target',
                    'help': 'Name of target variable in data'},
                   {'tag': '--depth',
                    'help': 'Composer parameter: max depth of the pipeline'},
                   {'tag': '--arity',
                    'help': 'Composer parameter: max arity of the pipeline nodes'},
                   {'tag': '--popsize',
                    'help': 'Composer parameter: population size'},
                   {'tag': '--gen_num',
                    'help': 'Composer parameter: number of generations'},
                   {'tag': '--opers',
                    'help': 'Composer parameter: model names to use',
                    'is_list': True},
                   {'tag': '--tuning',
                    'help': 'Composer parameter: 1 - with tuning, 0 - without tuning'},
                   {'tag': '--cv_folds',
                    'help': 'Composer parameter: Number of folds for cross-validation'},
                   {'tag': '--val_bl',
                    'help': 'Composer parameter: Number of validation blocks for time series forecasting'},
                   {'tag': '--hist_path',
                    'help': 'Composer parameter: Name of the folder for composing history'},
                   {'tag': '--for_len',
                    'help': 'Time Series Forecasting parameter: forecast length'},
                   ]
# dictionary with tag - name relation
keys_names = {'problem': 'problem',
              'preset': 'preset',
              'timeout': 'timeout',
              'seed': 'seed',
              'train': 'features',
              'target': 'target',
              'depth': 'max_depth',
              'arity': 'max_arity',
              'popsize': 'pop_size',
              'gen_num': 'num_of_generations',
              'opers': 'available_operations',
              'tuning': 'with_tuning',
              'cv_folds': 'cv_folds',
              'val_bl': 'validation_blocks',
              'hist_path': 'history_folder',
              'for_len': 'forecast_length'
              }


def run_cli():
    parser = create_parser(arguments_dicts)
    parameters = parser.parse_args()

    main_params, fit_params = separate_argparse_to_fedot(parameters)

    preprocess_keys(main_params)
    preprocess_keys(fit_params)

    run_fedot(parameters, main_params, fit_params)


if __name__ == '__main__':
    run_cli()



from functools import partial
from typing import Optional

from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.log import default_log
from fedot.core.repository.tasks import TaskTypesEnum
from fedot.core.validation.split import tabular_cv_generator, ts_cv_generator
from fedot.remote.remote_evaluator import RemoteEvaluator, init_data_for_remote_execution
from .data_objective_advisor import DataObjectiveAdvisor
from .data_objective_eval import DataSource
from ...constants import default_data_split_ratio_by_task


class DataSourceSplitter:
    """
    Splitter of data that provides generator of test-train splits.
    Can provide hold-out validation and k-fold validation.

    :param cv_folds: Number of folds on data for cross-validation.
    If provided, then k-fold validation is used. Otherwise, hold-out validation is used.
    :param validation_blocks: Number of validation blocks for time series forecasting.
    :param split_ratio: Ratio of data for splitting. Applied only in case of hold-out split.
    If not provided, then default split ratios will be used.
    :param shuffle: Is shuffling required for data.
    """

    def __init__(self,
                 cv_folds: Optional[int] = None,
                 validation_blocks: Optional[int] = None,
                 split_ratio: Optional[float] = None,
                 shuffle: bool = False):
        self.cv_folds = cv_folds
        self.validation_blocks = validation_blocks
        self.split_ratio = split_ratio
        self.shuffle = shuffle
        self.advisor = DataObjectiveAdvisor()
        self.log = default_log(self)

    def build(self, data: InputData) -> DataSource:
        # Shuffle data
        if self.shuffle and data.task.task_type is not TaskTypesEnum.ts_forecasting:
            data.shuffle()

        # Split data
        if self.cv_folds is not None:
            self.log.info("K-folds cross validation is applied.")
            data_producer = self._build_kfolds_producer(data)
        else:
            self.log.info("Hold out validation is applied.")
            data_producer = self._build_holdout_producer(data)

        return data_producer

    @staticmethod
    def _data_producer(train_data: InputData, test_data: InputData):
        yield train_data, test_data

    def _build_holdout_producer(self, data: InputData) -> DataSource:
        """
        Build trivial data producer for hold-out validation
        that always returns same data split. Equivalent to 1-fold validation.
        """

        split_ratio = self.split_ratio or default_data_split_ratio_by_task[data.task.task_type]
        train_data, test_data = train_test_data_setup(data, split_ratio, validation_blocks=self.validation_blocks)

        if RemoteEvaluator().is_enabled:
            init_data_for_remote_execution(train_data)

        return partial(self._data_producer, train_data, test_data)

    def _build_kfolds_producer(self, data: InputData) -> DataSource:
        if isinstance(data, MultiModalData):
            raise NotImplementedError('Cross-validation is not supported for multi-modal data')
        if data.task.task_type is TaskTypesEnum.ts_forecasting:
            # Perform time series cross validation
            if self.validation_blocks is None:
                default_validation_blocks = 2
                self.validation_blocks = default_validation_blocks
                self.log.info('For timeseries cross validation validation_blocks number was changed ' +
                              f'from None to {default_validation_blocks} blocks')
            cv_generator = partial(ts_cv_generator, data,
                                   self.cv_folds,
                                   self.validation_blocks,
                                   self.log)
        else:
            cv_generator = partial(tabular_cv_generator, data,
                                   self.cv_folds,
                                   self.advisor.propose_kfold(data))
        return cv_generator



import os
from typing import Tuple

from fedot.core.operations.operation_template import OperationTemplateAbstract, check_existing_path
from fedot.core.pipelines.node import Node


class AtomizedModelTemplate(OperationTemplateAbstract):
    def __init__(self, node: Node = None, operation_id: int = None, nodes_from: list = None, path: str = None):
        # Need use the imports inside the class because of the problem of circular imports.
        from fedot.core.pipelines.pipeline import Pipeline
        from fedot.core.pipelines.template import PipelineTemplate
        from fedot.core.operations.atomized_model import AtomizedModel

        super().__init__()
        self.atomized_model_json_path = None
        self.next_pipeline_template = None
        self.pipeline_template = None

        if path:
            pipeline = Pipeline.from_serialized(path)
            self.next_pipeline_template = AtomizedModel(pipeline)
            self.pipeline_template = PipelineTemplate(pipeline)

        if node:
            self._operation_to_template(node, operation_id, nodes_from)

    def _operation_to_template(self, node: Node, operation_id: int, nodes_from: list):
        from fedot.core.pipelines.template import PipelineTemplate

        self.operation_id = operation_id
        self.operation_type = node.operation.operation_type
        self.nodes_from = nodes_from
        self.pipeline_template = PipelineTemplate(node.operation.pipeline)
        self.atomized_model_json_path = 'nested_' + str(self.operation_id)

    def convert_to_dict(self) -> dict:

        operation_object = {
            'operation_id': self.operation_id,
            'operation_type': self.operation_type,
            'nodes_from': self.nodes_from,
            'atomized_model_json_path': self.atomized_model_json_path
        }

        return operation_object

    def _create_nested_path(self, path: str) -> Tuple[str, str]:
        """
        Create folder for nested JSON operation and prepared path to save JSON's.
        :params path: path where to save parent JSON operation
        :return: absolute and relative paths to save nested JSON operation
        """

        relative_path = os.path.join('fitted_operations', 'nested_' + str(self.operation_id))
        absolute_path = os.path.join(path, relative_path)

        if not os.path.exists(absolute_path):
            os.makedirs(absolute_path)

        return absolute_path, relative_path

    def export_operation(self, path: str):
        absolute_path = os.path.join(path, self.atomized_model_json_path)
        check_existing_path(absolute_path)
        self.pipeline_template.export_pipeline(absolute_path, create_subdir=False)

    def import_json(self, operation_object: dict):
        required_fields = ['operation_id', 'operation_type', 'nodes_from', 'atomized_model_json_path']
        self._validate_json_operation_template(operation_object, required_fields)

        self.operation_id = operation_object['operation_id']
        self.operation_type = operation_object['operation_type']
        self.nodes_from = operation_object['nodes_from']
        self.atomized_model_json_path = operation_object['atomized_model_json_path']



from fedot.core.operations.model import Model
from fedot.core.repository.operation_types_repository import OperationTypesRepository


class AutoML(Model):
    """
    Class with fit/predict methods defining the automl strategy for the task

    :param operation_type: name of the model
    """

    def __init__(self, operation_type: str):
        super().__init__(operation_type=operation_type)
        self.operations_repo = OperationTypesRepository('automl')



import os

from sklearn.metrics import roc_auc_score as roc_auc

from cases.dataset_preparation import unpack_archived_data
from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline


def execute_pipeline_for_text_problem(train_data, test_data):
    node_text_clean = PrimaryNode('text_clean')
    node_tfidf = SecondaryNode('tfidf', nodes_from=[node_text_clean])
    model_node = SecondaryNode('multinb', nodes_from=[node_tfidf])
    pipeline = Pipeline(model_node)
    pipeline.fit(train_data)

    predicted = pipeline.predict(test_data)

    roc_auc_metric = roc_auc(y_true=test_data.target, y_score=predicted.predict)

    return roc_auc_metric


def run_text_problem_from_meta_file():
    data_file_abspath = os.path.abspath(os.path.join('data', 'spam', 'spamham.csv'))

    data = InputData.from_text_meta_file(meta_file_path=data_file_abspath)

    train_data, test_data = train_test_data_setup(data, split_ratio=0.7)

    metric = execute_pipeline_for_text_problem(train_data, test_data)

    print(f'meta_file metric: {metric}')


def run_text_problem_from_files():
    data_abspath = os.path.abspath(os.path.join('data', 'spamham'))

    unpack_archived_data(data_abspath)

    train_path = os.path.join(data_abspath, 'train')
    test_path = os.path.join(data_abspath, 'test')

    train_data = InputData.from_text_files(files_path=train_path)
    test_data = InputData.from_text_files(files_path=test_path)

    metric = execute_pipeline_for_text_problem(train_data, test_data)

    print(f'origin files metric: {metric}')


def run_text_problem_from_saved_meta_file(path):
    data = InputData.from_text_meta_file(meta_file_path=path)

    train_data, test_data = train_test_data_setup(data, split_ratio=0.7)

    metric = execute_pipeline_for_text_problem(train_data, test_data)

    print(f'meta_file metric: {metric}')


if __name__ == '__main__':
    run_text_problem_from_meta_file()
    run_text_problem_from_files()



import warnings
from typing import Optional

from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.evaluation_interfaces import EvaluationStrategy, SkLearnEvaluationStrategy
from fedot.core.operations.evaluation.operation_implementations.data_operations.decompose \
    import DecomposerRegImplementation
from fedot.core.operations.evaluation.operation_implementations.data_operations.sklearn_filters \
    import IsolationForestRegImplementation
from fedot.core.operations.evaluation.operation_implementations. \
    data_operations.sklearn_filters import LinearRegRANSACImplementation, NonLinearRegRANSACImplementation
from fedot.core.operations.evaluation.operation_implementations. \
    data_operations.sklearn_selectors import LinearRegFSImplementation, NonLinearRegFSImplementation
from fedot.core.operations.evaluation.operation_implementations.models.knn import FedotKnnRegImplementation
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.utilities.random import RandomStateHandler

warnings.filterwarnings("ignore", category=UserWarning)


class SkLearnRegressionStrategy(SkLearnEvaluationStrategy):
    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        """
        Predict method for regression task for predict stage
        :param trained_operation: model object
        :param predict_data: data used for prediction
        :return:
        """
        prediction = trained_operation.predict(predict_data.features)
        converted = self._convert_to_output(prediction, predict_data)

        return converted


class FedotRegressionPreprocessingStrategy(EvaluationStrategy):
    """ Strategy for applying custom algorithms from FEDOT to preprocess data
    for regression task
    """

    __operations_by_types = {
        'ransac_lin_reg': LinearRegRANSACImplementation,
        'ransac_non_lin_reg': NonLinearRegRANSACImplementation,
        'rfe_lin_reg': LinearRegFSImplementation,
        'rfe_non_lin_reg': NonLinearRegFSImplementation,
        'decompose': DecomposerRegImplementation,
        'isolation_forest_reg': IsolationForestRegImplementation
    }

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        self.operation_impl = self._convert_to_operation(operation_type)
        super().__init__(operation_type, params)

    def fit(self, train_data: InputData):
        """
        This method is used for operation training with the data provided
        :param InputData train_data: data used for operation training
        :return: trained data operation
        """

        warnings.filterwarnings("ignore", category=RuntimeWarning)
        operation_implementation = self.operation_impl(self.params_for_fit)
        with RandomStateHandler():
            operation_implementation.fit(train_data)
        return operation_implementation

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        """
        Transform method for preprocessing for predict stage

        :param trained_operation: model object
        :param predict_data: data used for prediction
        :return:
        """
        prediction = trained_operation.transform(predict_data)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def predict_for_fit(self, trained_operation, predict_data: InputData) -> OutputData:
        """
        Transform method for preprocessing for fit stage

        :param trained_operation: model object
        :param predict_data: data used for prediction
        :return:
        """
        prediction = trained_operation.transform_for_fit(predict_data)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain Custom Regression Preprocessing Strategy for {operation_type}')


class FedotRegressionStrategy(EvaluationStrategy):
    """
    Strategy for applying custom regression models from FEDOT make predictions
    """

    __operations_by_types = {
        'knnreg': FedotKnnRegImplementation
    }

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        self.operation_impl = self._convert_to_operation(operation_type)
        super().__init__(operation_type, params)

    def fit(self, train_data: InputData):
        """ This method is used for operation training """

        warnings.filterwarnings("ignore", category=RuntimeWarning)
        operation_implementation = self.operation_impl(self.params_for_fit)
        with RandomStateHandler():
            operation_implementation.fit(train_data)
        return operation_implementation

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:

        prediction = trained_operation.predict(predict_data)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def predict_for_fit(self, trained_operation, predict_data: InputData) -> OutputData:

        prediction = trained_operation.predict_for_fit(predict_data)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain Fedot Regression Strategy for {operation_type}')



from abc import abstractmethod
from dataclasses import dataclass
from typing import Any, Callable, Optional, Sequence

from fedot.core.adapter import BaseOptimizationAdapter, IdentityAdapter
from fedot.core.dag.verification_rules import DEFAULT_DAG_RULES
from fedot.core.optimisers.advisor import DefaultChangeAdvisor
from fedot.core.dag.graph import Graph
from fedot.core.dag.graph_verifier import GraphVerifier, VerifierRuleType
from fedot.core.log import default_log
from fedot.core.optimisers.archive import GenerationKeeper
from fedot.core.optimisers.composer_requirements import ComposerRequirements
from fedot.core.optimisers.gp_comp.operators.operator import PopulationT
from fedot.core.optimisers.graph import OptGraph
from fedot.core.optimisers.objective import GraphFunction, Objective, ObjectiveFunction
from fedot.core.optimisers.opt_history_objects.opt_history import OptHistory
from fedot.core.optimisers.opt_node_factory import DefaultOptNodeFactory, OptNodeFactory
from fedot.core.optimisers.gp_comp.evaluation import DelegateEvaluator

OptimisationCallback = Callable[[PopulationT, GenerationKeeper], Any]


def do_nothing_callback(*args, **kwargs):
    pass


@dataclass
class GraphOptimizerParameters:
    """Base class for definition of optimizer parameters. Can be extended for custom optimizers.

    :param multi_objective: defines if the optimizer must be multi-criterial
    :param offspring_rate: offspring rate used on next population
    :param pop_size: initial population size
    :param max_pop_size: maximum population size; optional, if unspecified, then population size is unbound
    :param adaptive_depth: flag to enable adaptive configuration of graph depth
    :param adaptive_depth_max_stagnation: max number of stagnating populations before adaptive depth increment
    """

    multi_objective: bool = False
    offspring_rate: float = 0.5
    pop_size: int = 20
    max_pop_size: Optional[int] = 55
    adaptive_depth: bool = False
    adaptive_depth_max_stagnation: int = 3


@dataclass
class GraphGenerationParams:
    """
    This dataclass is for defining the parameters using in graph generation process

    :param adapter: the function for processing of external object that should be optimized
    :param rules_for_constraint: collection of constraints
    :param advisor: class of task-specific advices for graph changes
    :param node_factory: class of generating nodes while mutation
    """
    adapter: BaseOptimizationAdapter
    verifier: GraphVerifier
    advisor: DefaultChangeAdvisor
    node_factory: OptNodeFactory
    remote_evaluator: Optional[DelegateEvaluator] = None

    def __init__(self, adapter: Optional[BaseOptimizationAdapter] = None,
                 rules_for_constraint: Sequence[VerifierRuleType] = tuple(DEFAULT_DAG_RULES),
                 advisor: Optional[DefaultChangeAdvisor] = None,
                 node_factory: Optional[OptNodeFactory] = None,
                 available_node_types: Optional[Sequence[Any]] = None,
                 remote_evaluator: Optional[DelegateEvaluator] = None,
                 ):
        self.adapter = adapter or IdentityAdapter()
        self.verifier = GraphVerifier(rules_for_constraint, self.adapter)
        self.advisor = advisor or DefaultChangeAdvisor()
        self.remote_evaluator = remote_evaluator
        if node_factory:
            self.node_factory = node_factory
        elif available_node_types:
            self.node_factory = DefaultOptNodeFactory(available_node_types)
        else:
            raise AttributeError('Required arguments not provided, please provide '
                                 'custom node factory or collection of available node types.')


class GraphOptimizer:
    """
    Base class of graph optimizer. It allows to find the optimal solution using specified metric (one or several).
    To implement the specific optimisation method,
    the abstract method 'optimize' should be re-defined in the ancestor class
    (e.g.  PopulationalOptimizer, RandomSearchGraphOptimiser, etc).

    :param objective: objective for optimisation
    :param initial_graphs: graphs which were initialized outside the optimizer
    :param requirements: implementation-independent requirements for graph optimizer
    :param graph_generation_params: parameters for new graph generation
    :param graph_optimizer_parameters: parameters for specific implementation of graph optimizer
    """

    def __init__(self,
                 objective: Objective,
                 initial_graphs: Optional[Sequence[Graph]] = None,
                 requirements: Optional[ComposerRequirements] = None,
                 graph_generation_params: Optional[GraphGenerationParams] = None,
                 graph_optimizer_parameters: Optional[GraphOptimizerParameters] = None):
        self.log = default_log(self)
        self.initial_graphs = initial_graphs
        self._objective = objective
        self.requirements = requirements or ComposerRequirements()
        self.graph_generation_params = graph_generation_params
        self.graph_optimizer_params = graph_optimizer_parameters or GraphOptimizerParameters()
        self._optimisation_callback: OptimisationCallback = do_nothing_callback
        mo = False if not graph_optimizer_parameters else graph_optimizer_parameters.multi_objective
        self.history = OptHistory(mo, requirements.history_dir) \
            if requirements and requirements.keep_history else None

    @property
    def objective(self) -> Objective:
        return self._objective

    @abstractmethod
    def optimise(self, objective: ObjectiveFunction) -> Sequence[OptGraph]:
        """
        Method for running of optimization using specified algorithm.
        :param objective: objective function that specifies optimization target
        :return: sequence of the best graphs
        """
        pass

    def set_optimisation_callback(self, callback: Optional[OptimisationCallback]):
        """Set optimisation callback that must run on each iteration.
        Reset the callback if None is passed."""
        self._optimisation_callback = callback or do_nothing_callback

    def set_evaluation_callback(self, callback: Optional[GraphFunction]):
        """Set or reset (with None) post-evaluation callback
        that's called on each graph after its evaluation."""
        pass



from typing import List

from fedot.api.api_utils.assumptions.operations_filter import OperationsFilter
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.pipeline_builder import PipelineBuilder
from fedot.core.repository.operation_types_repository import OperationTypesRepository
from fedot.core.repository.tasks import Task, TaskTypesEnum


class TaskAssumptions:
    """ Abstracts task-specific pipeline assumptions. """

    def __init__(self, repository: OperationTypesRepository):
        self.repo = repository

    @staticmethod
    def for_task(task: Task, repository: OperationTypesRepository):
        assumptions_by_task = {
            TaskTypesEnum.classification: ClassificationAssumptions,
            TaskTypesEnum.regression: RegressionAssumptions,
            TaskTypesEnum.ts_forecasting: TSForecastingAssumptions,
        }
        assumptions_cls = assumptions_by_task.get(task.task_type)
        if not assumptions_cls:
            raise NotImplementedError(f"Don't have assumptions for task type: {task.task_type}")
        return assumptions_cls(repository)

    def ensemble_operation(self) -> str:
        """ Suitable ensemble operation used for MultiModalData case. """
        raise NotImplementedError()

    def processing_builders(self) -> List[PipelineBuilder]:
        """ Returns alternatives of PipelineBuilders for core processing (without preprocessing). """
        raise NotImplementedError()

    def fallback_builder(self, operations_filter: OperationsFilter) -> PipelineBuilder:
        """
        Returns default PipelineBuilder for case when primary alternatives are not valid.
        Have access for OperationsFilter for sampling available operations.
        """
        raise NotImplementedError()


class TSForecastingAssumptions(TaskAssumptions):
    """ Simple static dictionary-based assumptions for time series forecasting task. """

    @property
    def builders(self):
        return {
            'glm_ridge':
                PipelineBuilder()
                .add_branch('glm', 'lagged')
                .add_node('ridge', branch_idx=1)
                .join_branches('ridge'),
            'lagged_ridge':
                PipelineBuilder()
                .add_sequence('lagged', 'ridge'),
            'polyfit_ridge':
                PipelineBuilder()
                .add_branch('polyfit', 'lagged')
                .grow_branches(None, 'ridge')
                .join_branches('ridge'),
            'smoothing_ar':
                PipelineBuilder()
                .add_sequence('smoothing', 'ar'),
        }

    def ensemble_operation(self) -> str:
        return 'ridge'

    def processing_builders(self) -> List[Pipeline]:
        return list(self.builders.values())

    def fallback_builder(self, operations_filter: OperationsFilter) -> PipelineBuilder:
        random_choice_node = operations_filter.sample()
        operation_info = self.repo.operation_info_by_id(random_choice_node)
        if 'non_lagged' in operation_info.tags:
            return PipelineBuilder().add_node(random_choice_node)
        else:
            return PipelineBuilder().add_node('lagged').add_node(random_choice_node)


class RegressionAssumptions(TaskAssumptions):
    """ Simple static dictionary-based assumptions for regression task. """

    @property
    def builders(self):
        return {
            'rfr': PipelineBuilder().add_node('rfr'),
            'ridge': PipelineBuilder().add_node('ridge'),
        }

    def ensemble_operation(self) -> str:
        return 'rfr'

    def processing_builders(self) -> List[Pipeline]:
        return list(self.builders.values())

    def fallback_builder(self, operations_filter: OperationsFilter) -> PipelineBuilder:
        random_choice_node = operations_filter.sample()
        return PipelineBuilder().add_node(random_choice_node)


class ClassificationAssumptions(TaskAssumptions):
    """ Simple static dictionary-based assumptions for classification task. """

    @property
    def builders(self):
        return {
            'rf': PipelineBuilder().add_node('rf'),
            'logit': PipelineBuilder().add_node('logit'),
            'cnn': PipelineBuilder().add_node('cnn'),
        }

    def ensemble_operation(self) -> str:
        return 'rf'

    def processing_builders(self) -> List[Pipeline]:
        return list(self.builders.values())

    def fallback_builder(self, operations_filter: OperationsFilter) -> PipelineBuilder:
        random_choice_node = operations_filter.sample()
        return PipelineBuilder().add_node(random_choice_node)



import logging
from random import choice
from typing import Any, Optional, Sequence, Union

from fedot.api.main import Fedot
from fedot.core.composer.gp_composer.specific_operators import boosting_mutation, parameter_change_mutation
from fedot.core.dag.graph import Graph
from fedot.core.optimisers.gp_comp.evaluation import SimpleDispatcher
from fedot.core.optimisers.gp_comp.operators.mutation import MutationTypesEnum, Mutation
from fedot.core.optimisers.objective import Objective, ObjectiveFunction
from fedot.core.optimisers.opt_history_objects.individual import Individual
from fedot.core.optimisers.optimizer import GraphGenerationParams, GraphOptimizer, GraphOptimizerParameters
from fedot.core.optimisers.timer import OptimisationTimer
from fedot.core.utils import fedot_project_root


class RandomMutationSearchOptimizer(GraphOptimizer):
    """
    Random search-based graph models optimizer
    """

    def __init__(self,
                 objective: Objective,
                 initial_graphs: Union[Graph, Sequence[Graph]] = (),
                 requirements: Optional[Any] = None,
                 graph_generation_params: Optional[GraphGenerationParams] = None,
                 graph_optimizer_parameters: Optional[GraphOptimizerParameters] = None):
        super().__init__(objective, initial_graphs, requirements, graph_generation_params, graph_optimizer_parameters)
        self.mutation_types = [boosting_mutation, parameter_change_mutation,
                               MutationTypesEnum.single_edge,
                               MutationTypesEnum.single_change,
                               MutationTypesEnum.single_drop,
                               MutationTypesEnum.single_add]

    def optimise(self, objective: ObjectiveFunction):

        timer = OptimisationTimer(timeout=self.requirements.timeout)
        dispatcher = SimpleDispatcher(self.graph_generation_params.adapter)
        evaluator = dispatcher.dispatch(objective, timer)

        num_iter = 0

        initial_individuals = [Individual(graph) for graph in self.initial_graphs]
        best = choice(initial_individuals)
        evaluator([best])

        with timer as t:
            while not t.is_time_limit_reached(num_iter):
                mutation = Mutation(self.mutation_types, self.requirements, self.graph_generation_params)
                new = mutation(best)
                evaluator([new])
                if new.fitness < best.fitness:
                    best = new
                num_iter += 1

        return self.graph_generation_params.adapter.restore(best)


def run_with_random_search_composer():
    train_data_path = f'{fedot_project_root()}/cases/data/scoring/scoring_train.csv'
    test_data_path = f'{fedot_project_root()}/cases/data/scoring/scoring_test.csv'

    composer_params = {'available_operations': ['class_decompose', 'rf', 'linear', 'xgboost', 'dt'],
                       'optimizer': RandomMutationSearchOptimizer}

    automl = Fedot(problem='classification', timeout=1, logging_level=logging.DEBUG,
                   preset='fast_train', **composer_params)

    automl.fit(train_data_path)
    automl.predict(test_data_path)
    print(automl.get_metrics())


if __name__ == '__main__':
    run_with_random_search_composer()



from typing import List, Union, Optional, Set

from fedot.api.api_utils.assumptions.operations_filter import OperationsFilter, WhitelistOperationsFilter
from fedot.api.api_utils.assumptions.preprocessing_builder import PreprocessingBuilder
from fedot.api.api_utils.assumptions.task_assumptions import TaskAssumptions
from fedot.core.log import default_log
from fedot.core.data.data import InputData
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.operation_types_repository import OperationTypesRepository
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.pipeline_builder import PipelineBuilder, Node


class AssumptionsBuilder:

    def __init__(self, data: Union[InputData, MultiModalData], repository_name: str = 'model'):
        self.logger = default_log(prefix='FEDOT logger')
        self.data = data
        self.repo = OperationTypesRepository(repository_name)
        self.assumptions_generator = TaskAssumptions.for_task(self.data.task, self.repo)

    @staticmethod
    def get(data: Union[InputData, MultiModalData], repository_name: Optional[str] = None):
        if not repository_name:
            if data.data_type == DataTypesEnum.multi_ts:
                # It is needed to use data operations also for multi_ts data
                repository_name = 'all'
            else:
                repository_name = 'model'

        if isinstance(data, InputData):
            cls = UniModalAssumptionsBuilder
        elif isinstance(data, MultiModalData):
            cls = MultiModalAssumptionsBuilder
        else:
            raise NotImplementedError(f"Can't build assumptions for data type: {type(data).__name__}")
        return cls(data, repository_name=repository_name)

    def from_operations(self, available_operations: List[str]):
        raise NotImplementedError('abstract')

    def to_builders(self, initial_node: Optional[Node] = None) -> List[PipelineBuilder]:
        raise NotImplementedError('abstract')

    def build(self, initial_node: Optional[Node] = None) -> List[Pipeline]:
        return [builder.to_pipeline() for builder in self.to_builders(initial_node)]


class UniModalAssumptionsBuilder(AssumptionsBuilder):
    UNSUITABLE_AVAILABLE_OPERATIONS_MSG = "Unable to construct an initial assumption from the passed " \
                                          "available operations, default initial assumption will be used"

    def __init__(self, data: Union[InputData, MultiModalData],
                 data_type: DataTypesEnum = None, repository_name: str = "model"):
        """ Construct builder from task and data.
        :param data: data that will be passed to the pipeline
        :param data_type: allows specifying data_type of particular column for MultiModalData case
        """
        super().__init__(data, repository_name)
        self.data_type = data_type or data.data_type
        self.ops_filter = OperationsFilter()

    def from_operations(self, available_operations: Optional[List[str]] = None):
        if available_operations:
            operations_for_task_and_data = self.repo.suitable_operation(self.data.task.task_type, self.data_type)
            operations_to_choose_from = set(operations_for_task_and_data).intersection(available_operations)
            _check_operations_to_choose_from(self.data, self.data_type, operations_to_choose_from)
            if operations_to_choose_from:
                self.ops_filter = WhitelistOperationsFilter(available_operations, operations_to_choose_from)
            else:
                # Don't filter pipelines as we're not able to create
                # fallback pipelines without operations_to_choose_from.
                # So, leave default dumb ops_filter.
                self.logger.info(self.UNSUITABLE_AVAILABLE_OPERATIONS_MSG)
        return self

    def to_builders(self, initial_node: Optional[Node] = None) -> List[PipelineBuilder]:
        """ Return a list of valid builders satisfying internal
        OperationsFilter or a single fallback builder. """
        preprocessing = \
            PreprocessingBuilder.builder_for_data(self.data.task.task_type, self.data, initial_node)
        valid_builders = []
        for processing in self.assumptions_generator.processing_builders():
            candidate_builder = preprocessing.merge_with(processing)
            if self.ops_filter.satisfies(candidate_builder.to_pipeline()):
                valid_builders.append(candidate_builder)
        return valid_builders or [self.assumptions_generator.fallback_builder(self.ops_filter)]


class MultiModalAssumptionsBuilder(AssumptionsBuilder):
    def __init__(self, data: MultiModalData, repository_name: str = "model"):
        super().__init__(data, repository_name)
        _subbuilders = []
        for data_type, (data_source_name, values) in zip(self.data.data_type, self.data.items()):
            _subbuilders.append((data_source_name, UniModalAssumptionsBuilder(self.data, data_type)))
        self._subbuilders = tuple(_subbuilders)

    def from_operations(self, available_operations: Optional[List[str]] = None):
        for data_source, subbuilder in self._subbuilders:
            # Performs specific filter on image data operations
            if subbuilder.data_type is DataTypesEnum.image:
                available_img_operations = ['data_source_img', 'cnn']
                subbuilder.from_operations(available_img_operations)
        return self

    def to_builders(self, initial_node: Optional[Node] = None) -> List[PipelineBuilder]:
        # For each data source build its own list of alternatives of initial pipelines.
        subpipelines: List[List[Pipeline]] = []
        for data_source_name, subbuilder in self._subbuilders:
            first_node = PipelineBuilder().add_node(data_source_name).add_node(initial_node).to_nodes()[0]
            data_pipeline_alternatives = subbuilder.build(first_node)
            subpipelines.append(data_pipeline_alternatives)

        # Then zip these alternatives together and add final node to get ensembles.
        ensemble_builders: List[PipelineBuilder] = []
        for pre_ensemble in zip(*subpipelines):
            ensemble_operation = self.assumptions_generator.ensemble_operation()
            ensemble_nodes = map(lambda pipeline: pipeline.root_node, pre_ensemble)
            ensemble_builder = PipelineBuilder(*ensemble_nodes).join_branches(ensemble_operation)
            ensemble_builders.append(ensemble_builder)
        return ensemble_builders


def _check_operations_to_choose_from(data, data_type: DataTypesEnum, operations_to_choose_from: Set[str]):
    """Since it is sometimes impossible to form a valid pipeline without some operations,
     they are added to the set of operations for current task and data."""
    if isinstance(data, MultiModalData):
        if data_type is DataTypesEnum.image and 'data_source_img' not in operations_to_choose_from:
            operations_to_choose_from.add('data_source_img')
        if data_type is DataTypesEnum.text and 'data_source_text' not in operations_to_choose_from:
            operations_to_choose_from.add('data_source_text')
        if data_type is DataTypesEnum.table and 'data_source_table' not in operations_to_choose_from:
            operations_to_choose_from.add('data_source_table')
    if data_type is DataTypesEnum.image and 'cnn' not in operations_to_choose_from:
        operations_to_choose_from.add('cnn')



from fedot.core.operations.automl import AutoML
from fedot.core.operations.data_operation import DataOperation
from fedot.core.operations.model import Model
from fedot.core.operations.operation import Operation
from fedot.core.repository.operation_types_repository import OperationTypesRepository, get_operation_type_from_id


class OperationFactory:
    """
    Base class for determining what type of operations should be defined
    in the node. Possible operations are models (ML models with fit and predict
    methods) and data operations (e.g. scaling) with fit and transform methods

    """

    def __init__(self, operation_name):
        self.operation_name = operation_name
        self.operation_type = self._define_operation_type()

    def get_operation(self) -> Operation:
        """
        Factory method returns the desired object of the 'Data_operation' or
        'Model' class which depends on model_type variable

        """

        if self.operation_type == 'model':
            operation = Model(operation_type=self.operation_name)
        elif self.operation_type == 'data_operation':
            operation = DataOperation(operation_type=self.operation_name)
        elif self.operation_type == 'automl':
            operation = AutoML(operation_type=self.operation_name)
        else:
            raise ValueError(f'Operation type {self.operation_type} is not supported')

        return operation

    @property
    def operation_type_name(self):
        return self.operation_type

    def _define_operation_type(self) -> str:
        """
        The method determines what type of operations is set for this node

        :return : operations type 'model', 'automl' or 'data_operation'
        """

        # Get available models from model_repository.json file
        operations_repo = OperationTypesRepository('data_operation')
        operations = operations_repo.operations
        if 'automl' in OperationTypesRepository.get_available_repositories():
            automl_repo = OperationTypesRepository('automl')
            models_automl = automl_repo.operations
        else:
            models_automl = []

        operation_name = get_operation_type_from_id(self.operation_name)

        # If there is a such model in the list
        if any(operation_name == model.id for model in operations):
            operation_type = 'data_operation'
        elif any(operation_name == model.id for model in models_automl):
            operation_type = 'automl'
        # Otherwise - it is model
        else:
            operation_type = 'model'
        return operation_type



from fedot.core.repository.tasks import TaskTypesEnum
from fedot.explainability.surrogate_explainer import SurrogateExplainer


def pick_pipeline_explainer(pipeline: 'Pipeline', method: str, task_type: TaskTypesEnum):
    if method == 'surrogate_dt':
        if task_type == TaskTypesEnum.classification:
            surrogate = 'dt'
        elif task_type == TaskTypesEnum.regression:
            surrogate = 'dtreg'
        else:
            raise ValueError(f'Surrogate tree is not applicable for the {task_type} task')
        explainer = SurrogateExplainer(pipeline, surrogate=surrogate)

    else:
        raise ValueError(f'Explanation method {method} is not supported')

    return explainer


def explain_pipeline(pipeline: 'Pipeline', data: 'InputData', method: str = 'surrogate_dt',
                     visualization: bool = False, **kwargs) -> 'Explainer':
    """Create explanation for the `pipeline` according to the selected `method`.
    An `Explainer` instance is returned.

    :param pipeline: pipeline to explain.
    :param data: samples to be explained.
    :param method: explanation method, defaults to 'surrogate_dt'. Options: ['surrogate_dt', ...]
    :param visualization: print and plot the explanation simultaneously, defaults to True.
        The explanation can be retrieved later by executing `explainer.visualize()`.
    """
    if not pipeline:
        raise AssertionError('The pipeline might be fit before explanation!')

    explainer = pick_pipeline_explainer(pipeline, method, data.task.task_type)
    explainer.explain(data, visualization=visualization, **kwargs)

    return explainer



import numpy as np

from fedot.core.data.data import OutputData
from fedot.core.operations.operation import Operation
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.operation_types_repository import OperationTypesRepository
from fedot.core.repository.tasks import TaskTypesEnum


class Model(Operation):
    """Class with ``fit``/``predict`` methods defining the evaluation strategy for the task

    Args:
        operation_type: name of the model
    """

    def __init__(self, operation_type: str):
        super().__init__(operation_type=operation_type)
        self.operations_repo = OperationTypesRepository('model')

    @staticmethod
    def assign_tabular_column_types(output_data: OutputData, output_mode: str) -> OutputData:
        """Assign types for tabular data obtained from model predictions.\n
        By default, all types of model predictions for tabular data can be clearly defined
        """
        if output_data.data_type is not DataTypesEnum.table:
            # No column data types info for non-tabular data
            return output_data

        is_regression_task = output_data.task.task_type is TaskTypesEnum.regression
        is_ts_forecasting_task = output_data.task.task_type is TaskTypesEnum.ts_forecasting

        predict_shape = np.array(output_data.predict).shape
        # Add information about features
        if is_regression_task or is_ts_forecasting_task:
            if len(predict_shape) < 2:
                column_info = {'features': [str(float)] * predict_shape[0]}
            else:
                column_info = {'features': [str(float)] * predict_shape[1]}
        else:
            if len(predict_shape) < 2:
                output_data.predict = output_data.predict.reshape((-1, 1))
                predict_shape = output_data.predict.shape
            # Classification task or clustering
            if output_mode == 'labels':
                column_info = {'features': [str(int)] * predict_shape[1]}
            else:
                column_info = {'features': [str(float)] * predict_shape[1]}

        # Add information about target
        target_shape = output_data.target.shape if output_data.target is not None else None
        if target_shape is None:
            # There is no target column in output data
            output_data.supplementary_data.column_types = column_info
            return output_data

        if is_regression_task or is_ts_forecasting_task:
            if len(target_shape) > 1:
                column_info.update({'target': [str(float)] * target_shape[1]})
            else:
                # Array present "time series"
                column_info.update({'target': [str(float)] * len(output_data.target)})
        else:
            # Classification task or clustering
            column_info.update({'target': [str(int)] * predict_shape[1]})

        output_data.supplementary_data.column_types = column_info
        return output_data



from os import makedirs
from os.path import exists, join

from examples.simple.classification.classification_pipelines import classification_three_depth_manual_pipeline
from examples.simple.regression.regression_pipelines import regression_three_depth_manual_pipeline
from fedot.core.composer.composer_builder import ComposerBuilder
from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.optimisers.gp_comp.gp_params import GPGraphOptimizerParameters
from fedot.core.optimisers.gp_comp.operators.inheritance import GeneticSchemeTypesEnum
from fedot.core.optimisers.gp_comp.pipeline_composer_requirements import PipelineComposerRequirements
from fedot.core.repository.operation_types_repository import get_operations_for_task
from fedot.core.repository.quality_metrics_repository import ClassificationMetricsEnum, MetricsRepository, \
    RegressionMetricsEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum
from fedot.core.utils import default_fedot_data_dir, fedot_project_root
from fedot.sensitivity.node_sa_approaches import NodeDeletionAnalyze, NodeReplaceOperationAnalyze
from fedot.sensitivity.nodes_sensitivity import NodesAnalysis


def get_composed_pipeline(dataset_to_compose, task, metric_function):
    # the search of the models provided by the framework that can be used as nodes in a pipeline for the selected task
    available_model_types = get_operations_for_task(task=task, mode='model')

    # the choice and initialisation of the GP search
    composer_requirements = PipelineComposerRequirements(
        primary=available_model_types,
        secondary=available_model_types,
        max_arity=3, max_depth=3,
        num_of_generations=20,
    )

    optimizer_parameters = GPGraphOptimizerParameters(
        pop_size=15,
        mutation_prob=0.8, crossover_prob=0.8,
        genetic_scheme_type=GeneticSchemeTypesEnum.steady_state,
    )

    # Create composer and with required composer params
    composer = ComposerBuilder(task=task). \
        with_requirements(composer_requirements). \
        with_optimizer_params(optimizer_parameters). \
        with_metrics(metric_function). \
        build()

    # the optimal pipeline generation by composition - the most time-consuming task
    pipeline_evo_composed = composer.compose_pipeline(data=dataset_to_compose)

    return pipeline_evo_composed


def get_scoring_data():
    file_path_train = 'cases/data/scoring/scoring_train.csv'
    full_path_train = join(str(fedot_project_root()), file_path_train)

    # a dataset for a final validation of the composed model
    file_path_test = 'cases/data/scoring/scoring_test.csv'
    full_path_test = join(str(fedot_project_root()), file_path_test)
    task = Task(TaskTypesEnum.classification)
    train = InputData.from_csv(full_path_train, task=task)
    test = InputData.from_csv(full_path_test, task=task)

    return train, test


def get_kc2_data():
    file_path = 'cases/data/kc2/kc2.csv'
    full_path = join(str(fedot_project_root()), file_path)
    task = Task(TaskTypesEnum.classification)
    data = InputData.from_csv(full_path, task=task)
    train, test = train_test_data_setup(data)

    return train, test


def get_cholesterol_data():
    file_path = 'cases/data/cholesterol/cholesterol.csv'
    full_path = join(str(fedot_project_root()), file_path)
    task = Task(TaskTypesEnum.regression)
    data = InputData.from_csv(full_path, task=task)
    train, test = train_test_data_setup(data)

    return train, test


def pipeline_by_task(task, metric, data, is_composed):
    if is_composed:
        pipeline = get_composed_pipeline(data, task,
                                         metric_function=metric)
    else:
        if task.task_type.name == 'classification':
            pipeline = classification_three_depth_manual_pipeline()
        else:
            pipeline = regression_three_depth_manual_pipeline()

    return pipeline


def run_analysis_case(train_data: InputData, test_data: InputData,
                      case_name: str, task, metric, is_composed=False, result_path=None):
    pipeline = pipeline_by_task(task=task, metric=metric,
                                data=train_data, is_composed=is_composed)

    pipeline.fit(train_data)

    if not result_path:
        result_path = join(default_fedot_data_dir(), 'sensitivity', f'{case_name}')
        if not exists(result_path):
            makedirs(result_path)

    pipeline.show(save_path=result_path)

    pipeline_analysis_result = NodesAnalysis(pipeline=pipeline, train_data=train_data,
                                             test_data=test_data, path_to_save=result_path,
                                             approaches=[NodeDeletionAnalyze,
                                                         NodeReplaceOperationAnalyze]).analyze()

    print(f'pipeline analysis result {pipeline_analysis_result}')


def run_class_scoring_case(is_composed: bool, path_to_save=None):
    train_data, test_data = get_scoring_data()
    task = Task(TaskTypesEnum.classification)
    # the choice of the metric for the pipeline quality assessment during composition
    metric_function = MetricsRepository().metric_by_id(ClassificationMetricsEnum.ROCAUC_penalty)

    if is_composed:
        case = 'scoring_composed'
        run_analysis_case(train_data, test_data, case, task,
                          metric=metric_function,
                          is_composed=True, result_path=path_to_save)
    else:
        case = 'scoring'
        run_analysis_case(train_data, test_data, case, task,
                          metric=metric_function,
                          is_composed=False, result_path=path_to_save)


def run_class_kc2_case(is_composed: bool = False, path_to_save=None):
    train_data, test_data = get_kc2_data()
    task = Task(TaskTypesEnum.classification)
    # the choice of the metric for the pipeline quality assessment during composition
    metric_function = MetricsRepository().metric_by_id(ClassificationMetricsEnum.ROCAUC_penalty)

    if is_composed:
        case = 'kc2_composed'
        run_analysis_case(train_data, test_data, case, task,
                          metric=metric_function,
                          is_composed=True, result_path=path_to_save)
    else:
        case = 'kc2'
        run_analysis_case(train_data, test_data, case, task,
                          metric=metric_function,
                          is_composed=False, result_path=path_to_save)


def run_regr_case(is_composed: bool = False, path_to_save=None):
    train_data, test_data = get_cholesterol_data()
    task = Task(TaskTypesEnum.regression)
    # the choice of the metric for the pipeline quality assessment during composition
    metric_function = MetricsRepository().metric_by_id(RegressionMetricsEnum.RMSE)

    if is_composed:
        case = 'cholesterol_composed'
        run_analysis_case(train_data, test_data, case, task,
                          metric=metric_function,
                          is_composed=True, result_path=path_to_save)
    else:
        case = 'cholesterol'
        run_analysis_case(train_data, test_data, case, task,
                          metric=metric_function,
                          is_composed=False, result_path=path_to_save)


if __name__ == '__main__':
    # scoring case manual
    run_class_scoring_case(is_composed=False)

    # kc2 case manual
    run_class_kc2_case(is_composed=False)

    # cholesterol regr case
    run_regr_case(is_composed=False)
